# SPDX-FileCopyrightText: 2024 Shell Global Solutions International B.V. All Rights Reserved.
#
# SPDX-License-Identifier: Apache-2.0

# -*- coding: utf-8 -*-
"""SourceMap module.

The class for the source maps used in pyELQ

"""
from dataclasses import dataclass, field
from typing import Union

import numpy as np

from pyelq.coordinate_system import Coordinate, make_latin_hypercube
from pyelq.sensor.sensor import Sensor


@dataclass
class SourceMap:
    """Defines SourceMap class.

    Attributes:
        location (Coordinate, optional): Coordinate object specifying the potential source locations
        prior_value (np.ndarray, optional): Array with prior values for each source
        inclusion_idx (np.ndarray, optional): Array of lists containing indices of the observations of a
            corresponding sensor_object which are within the inclusion_radius of that particular source
        inclusion_n_obs (list, optional): Array containing number of observations of a sensor_object within
            radius for each source

    """

    location: Coordinate = field(init=False, default=None)
    prior_value: np.ndarray = None
    inclusion_idx: np.ndarray = field(init=False, default=None)
    inclusion_n_obs: np.ndarray = field(init=False, default=None)

    @property
    def nof_sources(self) -> int:
        """Number of sources."""
        if self.location is None:
            return 0
        return self.location.nof_observations

    def calculate_inclusion_idx(self, sensor_object: Sensor, inclusion_radius: Union[int, np.ndarray]) -> None:
        """Find observation indices which are within specified radius of each source location.

        This method takes the sensor object and for each source in the source_map object it calculates which
        observations are within the specified radius.
        When sensor_object location and sourcemap_object location are not of the same type, simply convert both to ECEF
        and calculate inclusion indices accordingly.
        The result is an array of lists which are the indices of the observations in sensor_object which are within the
        specified radius. Result is stored in the corresponding attribute.
        Also calculating number of observations in radius per source and storing result as a list in inclusion_n_obs
        attribute
        When a location attribute is in LLA we convert to ECEF for the inclusion radius to make sense

        Args:
            sensor_object (Sensor): Sensor object containing location information on the observations under
                consideration
            inclusion_radius (Union[float, np.ndarray], optional): Inclusion radius in [m] radius from source
                for which we take observations into account

        """
        sensor_kd_tree = sensor_object.location.to_ecef().create_tree()
        source_points = self.location.to_ecef().to_array()

        inclusion_idx = sensor_kd_tree.query_ball_point(source_points, inclusion_radius)
        idx_array = np.array(inclusion_idx, dtype=object)
        self.inclusion_idx = idx_array
        self.inclusion_n_obs = np.array([len(value) for value in self.inclusion_idx])

    def generate_sources(
        self,
        coordinate_object: Coordinate,
        sourcemap_limits: np.ndarray,
        sourcemap_type: str = "central",
        nof_sources: int = 5,
        grid_shape: Union[tuple, np.ndarray] = (5, 5, 1),
    ) -> None:
        """Generates source locations based on specified inputs.

        The result gets stored in the location attribute

        In grid_sphere we scale the latitude and longitude from -90/90 and -180/180 to 0/1 for the use in temp_lat_rad
        and temp_lon_rad

        Args:
            coordinate_object (Coordinate): Empty coordinate object which specifies the coordinate class to populate
                location with
            sourcemap_limits (np.ndarray): Limits of the sourcemap on which to generate the sources of size [dim x 2]
                if dim == 2 we assume the third dimension will be zeros. Assuming the units of the limits are defined in
                the desired coordinate system
            sourcemap_type (str, optional): Type of sourcemap to generate: central == 1 central source,
                hypercube == nof_sources through a Latin Hypercube design, grid == grid of shape grid_shape
                filled with sources, grid_sphere == grid of shape grid_shape taking into account a spherical spacing
            nof_sources (int, optional): Number of sources to generate (used in 'hypercube' case)
            grid_shape: (tuple, optional): Number of sources to generate in each dimension, total number of
                sources will be the product of the entries of this tuple (used in 'grid' and 'grid_sphere' case)

        """
        sourcemap_dimension = sourcemap_limits.shape[0]
        if sourcemap_type == "central":
            array = sourcemap_limits.mean(axis=1).reshape(1, sourcemap_dimension)
        elif sourcemap_type == "hypercube":
            array = make_latin_hypercube(bounds=sourcemap_limits, nof_samples=nof_sources)
        elif sourcemap_type == "grid":
            array = coordinate_object.make_grid(bounds=sourcemap_limits, grid_type="rectangular", shape=grid_shape)
        elif sourcemap_type == "grid_sphere":
            array = coordinate_object.make_grid(bounds=sourcemap_limits, grid_type="spherical", shape=grid_shape)
        else:
            raise NotImplementedError("Please provide a valid sourcemap type")
        coordinate_object.from_array(array=array)
        self.location = coordinate_object



# SPDX-FileCopyrightText: 2024 Shell Global Solutions International B.V. All Rights Reserved.
#
# SPDX-License-Identifier: Apache-2.0

# -*- coding: utf-8 -*-
"""Data access module.

Superclass containing some common attributes and helper functions used in multiple data access classes

"""

import datetime as dt
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Union

import pandas as pd

from pyelq.meteorology import Meteorology, MeteorologyGroup
from pyelq.sensor.sensor import Sensor, SensorGroup


@dataclass
class DataAccess(ABC):
    """DataAccess superclass containing some common attributes and functionalities.

    This superclass is used to show the type of methods to implement when creating a new data access class. The data
    access classes are used to convert raw data into well-defined classes and objects which can be used by the rest of
    the package.

    Attributes:
        latitude_bounds (tuple, optional): Tuple specifying (latitude_min, latitude_max)
        longitude_bounds (tuple, optional): Tuple specifying (longitude_min, longitude_max)
        date_bounds (tuple, optional): Tuple specifying (datetime_min, datetime_max)

    """

    latitude_bounds: tuple = (None, None)
    longitude_bounds: tuple = (None, None)
    date_bounds: tuple = (None, None)

    @abstractmethod
    def to_sensor(self, *args: Any, **kwargs: dict) -> Union[Sensor, SensorGroup]:
        """Abstract method to convert raw data into a Sensor or SensorGroup object.

        This method should be implemented to convert the raw data into a Sensor or SensorGroup object.

        Args:
            *args (Any): Variable length argument list of any type.
            **kwargs (dict): Arbitrary keyword arguments

        """

    @abstractmethod
    def to_meteorology(self, *args: Any, **kwargs: dict) -> Union[Meteorology, MeteorologyGroup]:
        """Abstract method to convert raw data into a Meteorology or MeteorologyGroup object.

        This method should be implemented to convert the raw data into a Meteorology or MeteorologyGroup object.

        Args:
            *args (Any): Variable length argument list of any type.
            **kwargs (dict): Arbitrary keyword arguments

        """

    def _query_aoi(self, data: pd.DataFrame) -> pd.DataFrame:
        """Helper function to perform area of interest query on data.

        Args:
            data (pd.Dataframe): Pandas dataframe to perform the query on

        """
        aoi_query_string = ""
        if self.latitude_bounds[0] is not None:
            aoi_query_string += f" & latitude>={self.latitude_bounds[0]}"
        if self.latitude_bounds[1] is not None:
            aoi_query_string += f" & latitude<={self.latitude_bounds[1]}"
        if self.longitude_bounds[0] is not None:
            aoi_query_string += f" & longitude>={self.longitude_bounds[0]}"
        if self.longitude_bounds[1] is not None:
            aoi_query_string += f" & longitude<={self.longitude_bounds[1]}"
        if len(aoi_query_string) > 0:
            aoi_query_string = aoi_query_string[3:]
            return data.query(aoi_query_string).copy()
        return data

    def _query_time(self, data: pd.DataFrame) -> pd.DataFrame:
        """Helper function to perform time query on data.

        Args:
            data (pd.Dataframe): Pandas dataframe to perform the query on

        """
        time_query_string = ""
        if self.date_bounds[0] is not None:
            timestamp_min = dt.datetime.timestamp(self.date_bounds[0])
            time_query_string += f" & timestamp>={timestamp_min}"
        if self.date_bounds[1] is not None:
            timestamp_max = dt.datetime.timestamp(self.date_bounds[1])
            time_query_string += f" & timestamp<={timestamp_max}"
        if len(time_query_string) > 0:
            time_query_string = time_query_string[3:]
            return data.query(time_query_string).copy()
        return data



# SPDX-FileCopyrightText: 2024 Shell Global Solutions International B.V. All Rights Reserved.
#
# SPDX-License-Identifier: Apache-2.0

# -*- coding: utf-8 -*-
"""Plot module.

Large module containing all the plotting code used to create various plots. Contains helper functions and the Plot class
definition.

"""
import re
import warnings
from copy import deepcopy
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Callable, Type, Union

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.figure_factory as ff
import plotly.graph_objects as go
from geojson import Feature, FeatureCollection
from openmcmc.mcmc import MCMC
from shapely import geometry

from pyelq.component.background import TemporalBackground
from pyelq.component.error_model import ErrorModel
from pyelq.component.offset import PerSensor
from pyelq.component.source_model import SlabAndSpike, SourceModel
from pyelq.coordinate_system import LLA
from pyelq.dispersion_model.gaussian_plume import GaussianPlume
from pyelq.sensor.sensor import Sensor, SensorGroup
from pyelq.support_functions.post_processing import (
    calculate_rectangular_statistics,
    create_lla_polygons_from_xy_points,
    is_regularly_spaced,
)

if TYPE_CHECKING:
    from pyelq.model import ELQModel

RGB_LIGHT_BLUE = "rgb(102, 197, 204)"
MCMC_ITERATION_NUMBER_LITERAL = "MCMC Iteration Number"


def lighter_rgb(rbg_string: str) -> str:
    """Takes in an RGB string and returns a lighter version of this colour.

    The colour is made lighter by increasing the magnitude of the RGB values by half of the difference between the
    original value and the number 255.

    Arguments:
        rbg_string (str): An RGB string.

    """
    rbg_string = rbg_string[4:-1]
    rbg_string = rbg_string.replace(" ", "")
    colors = rbg_string.split(",")
    colors_out = [np.nan, np.nan, np.nan]

    for i, color in enumerate(colors):
        color = int(color)
        color = min(int(round(color + ((255 - color) * 0.5))), 255)
        colors_out[i] = color

    return f"rgb({colors_out[0]}, {colors_out[1]}, {colors_out[2]})"


def plot_quantiles_from_array(
    fig: go.Figure,
    x_values: Union[np.ndarray, pd.arrays.DatetimeArray],
    y_values: np.ndarray,
    quantiles: Union[tuple, list, np.ndarray],
    color: str,
    name: str = None,
) -> go.Figure:
    """Plot quantiles over y-values against x-values.

    Assuming x-values have size N and y-values have size [N x M] where the second dimension is the dimension to
    calculate the quantiles over.

    Will plot the median of the y-values as a solid line and a filled area between the lower and upper specified
    quantile.

    Args:
        fig (go.Figure): Plotly figure to add the traces on.
        x_values (Union[np.ndarray, pd.arrays.DatetimeArray]): Numpy array containing the x-values to plot.
        y_values (np.ndarray): Numpy array containing the y-values to calculate the quantiles for.
        quantiles (Union[tuple, list, np.ndarray]): Values of upper and lower quantile to plot in range (0-100)
        color (str): RGB string specifying color for quantile fill plot.
        name (str, optional): Optional string name to show in the legend.

    Returns:
         fig (go.Figure): Plotly figure with the quantile filled traces and median trace added on it.

    """
    color_fill = f"rgba{color[3:-1]}, 0.3)"

    median_trace = go.Scatter(
        x=x_values,
        y=np.median(y_values, axis=1),
        mode="lines",
        line={"width": 3, "color": color},
        name=f"Median for {name}",
        legendgroup=name,
        showlegend=False,
    )

    lower_quantile_trace = go.Scatter(
        x=x_values,
        y=np.quantile(y_values, axis=1, q=quantiles[0] / 100),
        mode="lines",
        line={"width": 0, "color": color_fill},
        name=f"{quantiles[0]}% quantile",
        legendgroup=name,
        showlegend=False,
    )

    upper_quantile_trace = go.Scatter(
        x=x_values,
        y=np.quantile(y_values, axis=1, q=quantiles[1] / 100),
        fill="tonexty",
        fillcolor=color_fill,
        mode="lines",
        line={"width": 0, "color": color_fill},
        name=f"{quantiles[1]}% quantile",
        legendgroup=name,
        showlegend=False,
    )

    fig.add_trace(median_trace)
    fig.add_trace(lower_quantile_trace)
    fig.add_trace(upper_quantile_trace)

    return fig


def create_trace_specifics(object_to_plot: Union[Type[SlabAndSpike], SourceModel, MCMC], **kwargs: Any) -> dict:
    """Specification of different traces of single variables.

    Provides all details for plots where we want to plot a single variable as a line plot. Based on the object_to_plot
    we select the correct plot to show.

    Args:
        object_to_plot (Union[Type[SlabAndSpike], SourceModel, MCMC]): Object which we want to plot a single
            variable from
        **kwargs (Any): Additional key word arguments, e.g. burn_in or dict_key, used in some specific plots but not
            applicable to all.

    Returns:
        dict: A dictionary with the following key/values:
            x_values (Union[np.ndarray, pd.arrays.DatetimeArray]): Array containing the x-values to plot.
            y_values (np.ndarray): Numpy array containing the y-values to use in plotting.
            dict_key (str): String key associated with this plot to be used in the figure_dict attribute of the Plot
                class.
            title_text (str): String title of the plot.
            x_label (str): String label of x-axis.
            y_label (str) : String label of y-axis.
            name (str): String name to show in the legend.
            color (str): RGB string specifying color for plot.

    Raises:
        ValueError: When no specifics are defined for the inputted object to plot.

    """
    if isinstance(object_to_plot, SourceModel):
        dict_key = kwargs.pop("dict_key", "number_of_sources_plot")
        title_text = "Number of Sources 'on' against MCMC iterations"
        x_label = MCMC_ITERATION_NUMBER_LITERAL
        y_label = "Number of Sources 'on'"
        y_values = object_to_plot.number_on_sources
        x_values = np.array(range(y_values.size))
        color = "rgb(248, 156, 116)"
        name = "Number of Sources 'on'"

    elif isinstance(object_to_plot, MCMC):
        dict_key = kwargs.pop("dict_key", "log_posterior_plot")
        title_text = "Log posterior values against MCMC iterations"
        x_label = MCMC_ITERATION_NUMBER_LITERAL
        y_label = "Log Posterior<br>Value"
        y_values = object_to_plot.store["log_post"].flatten()
        x_values = np.array(range(y_values.size))
        color = RGB_LIGHT_BLUE
        name = "Log Posterior"

        if "burn_in" not in kwargs:
            warnings.warn("Burn in is not specified for the Log Posterior plot, are you sure this is correct?")

    else:
        raise ValueError("No values to plot")

    return {
        "x_values": x_values,
        "y_values": y_values,
        "dict_key": dict_key,
        "title_text": title_text,
        "x_label": x_label,
        "y_label": y_label,
        "name": name,
        "color": color,
    }


def create_plot_specifics(
    object_to_plot: Union[ErrorModel, PerSensor, MCMC], sensor_object: SensorGroup, plot_type: str = "", **kwargs: Any
) -> dict:
    """Specification of different traces where we want to plot a trace for each sensor.

    Provides all details for plots where we want to plot a single variable for each sensor as a line or box plot.
    Based on the object_to_plot we select the correct plot to show.

    When plotting the MCMC Observations and Predicted Model Values Against Time plot we are assuming time axis is the
    same for all sensors w.r.t. the fitted values from the MCMC store attribute, so we are only using the time axis
    from the first sensor.

    Args:
        object_to_plot (Union[ErrorModel, PerSensor, MCMC]): Object which we want to plot a single variable from
        sensor_object (SensorGroup): SensorGroup object associated with the object_to_plot
        plot_type (str, optional): String specifying either a line or a box plot.
        **kwargs (Any): Additional key word arguments, e.g. burn_in or dict_key, used in some specific plots but not
            applicable to all.

    Returns:
        dict: A dictionary with the following key/values:
            x_values (Union[np.ndarray, pd.arrays.DatetimeArray]): Array containing the x-values to plot.
            y_values (np.ndarray): Numpy array containing the y-values to use in plotting.
            dict_key (str): String key associated with this plot to be used in the figure_dict attribute of the
                Plot class.
            title_text (str): String title of the plot.
            x_label (str): String label of x-axis.
            y_label (str): String label of y-axis.
            plot_type (str): Type of plot which needs to be generated.

    Raises:
        ValueError: When no specifics are defined for the inputted object to plot.

    """
    if isinstance(object_to_plot, ErrorModel):
        y_values = np.sqrt(1 / object_to_plot.precision)
        x_values = np.array(range(y_values.shape[1]))

        if plot_type == "line":
            dict_key = kwargs.pop("dict_key", "error_model_iterations")
            title_text = "Estimated Error Model Values"
            x_label = MCMC_ITERATION_NUMBER_LITERAL
            y_label = "Estimated Error Model<br>Standard Deviation (ppm)"

        elif plot_type == "box":
            dict_key = kwargs.pop("dict_key", "error_model_distributions")
            title_text = "Distributions of Estimated Error Model Values After Burn-In"
            x_label = "Sensor"
            y_label = "Estimated Error Model<br>Standard Deviation (ppm)"

        else:
            raise ValueError("Only line and box are allowed for the plot_type argument for ErrorModel")

        if "burn_in" not in kwargs:
            warnings.warn("Burn in is not specified for the ErrorModel plot, are you sure this is correct?")

    elif isinstance(object_to_plot, PerSensor):
        offset_sensor_name = list(sensor_object.values())[0].label
        y_values = object_to_plot.offset
        nan_row = np.tile(np.nan, (1, y_values.shape[1]))
        y_values = np.concatenate((nan_row, y_values), axis=0)
        x_values = np.array(range(y_values.shape[1]))

        if plot_type == "line":
            dict_key = kwargs.pop("dict_key", "offset_iterations")
            title_text = f"Estimated Value of Offset w.r.t. {offset_sensor_name}"
            x_label = MCMC_ITERATION_NUMBER_LITERAL
            y_label = "Estimated Offset<br>Value (ppm)"

        elif plot_type == "box":
            dict_key = kwargs.pop("dict_key", "offset_distributions")
            title_text = f"Distributions of Estimated Offset Values w.r.t. {offset_sensor_name} After Burn-In"
            x_label = "Sensor"
            y_label = "Estimated Offset<br>Value (ppm)"

        else:
            raise ValueError("Only line and box are allowed for the plot_type argument for PerSensor OffsetModel")

        if "burn_in" not in kwargs:
            warnings.warn("Burn in is not specified for the PerSensor OffsetModel plot, are you sure this is correct?")

    elif isinstance(object_to_plot, MCMC):
        y_values = object_to_plot.store["y"]
        x_values = list(sensor_object.values())[0].time
        dict_key = kwargs.pop("dict_key", "fitted_values")
        title_text = "Observations and Predicted Model Values Against Time"
        x_label = "Time"
        y_label = "Concentration (ppm)"
        plot_type = "line"

    else:
        raise ValueError("No values to plot")

    return {
        "x_values": x_values,
        "y_values": y_values,
        "dict_key": dict_key,
        "title_text": title_text,
        "x_label": x_label,
        "y_label": y_label,
        "plot_type": plot_type,
    }


def plot_single_scatter(
    fig: go.Figure,
    x_values: Union[np.ndarray, pd.arrays.DatetimeArray],
    y_values: np.ndarray,
    color: str,
    name: str,
    **kwargs: Any,
) -> go.Figure:
    """Plots a single scatter trace on the supplied figure object.

    Args:
        fig (go.Figure): Plotly figure to add the trace to.
        x_values (Union[np.ndarray, pd.arrays.DatetimeArray]): X values to plot
        y_values (np.ndarray): Numpy array containing the y-values to use in plotting.
        color (str): RGB color string to use for this trace.
        name (str): String name to show in the legend.
        **kwargs (Any): Additional key word arguments, e.g. burn_in, legend_group, show_legend, used in some specific
            plots but not applicable to all.

    Returns:
        fig (go.Figure): Plotly figure with the trace added to it.

    """
    burn_in = kwargs.pop("burn_in", 0)
    legend_group = kwargs.pop("legend_group", name)
    show_legend = kwargs.pop("show_legend", True)
    if burn_in > 0:
        fig.add_trace(
            go.Scatter(
                x=x_values[: burn_in + 1],
                y=y_values[: burn_in + 1],
                name=name,
                mode="lines",
                line={"width": 3, "color": lighter_rgb(color)},
                legendgroup=legend_group,
                showlegend=False,
            )
        )

    fig.add_trace(
        go.Scatter(
            x=x_values[burn_in:],
            y=y_values[burn_in:],
            name=name,
            mode="lines",
            line={"width": 3, "color": color},
            legendgroup=legend_group,
            showlegend=show_legend,
        )
    )

    return fig


def plot_single_box(fig: go.Figure, y_values: np.ndarray, color: str, name: str) -> go.Figure:
    """Plot a single box plot trace on the plot figure.

    Args:
        fig (go.Figure): Plotly figure to add the trace to.
        y_values (np.ndarray): Numpy array containing the y-values to use in plotting.
        color (str): RGB color string to use for this trace.
        name (str): String name to show in the legend.

    Returns:
        fig (go.Figure): Plotly figure with the trace added to it.

    """
    fig.add_trace(go.Box(y=y_values, name=name, legendgroup=name, marker={"color": color}))

    return fig


def plot_polygons_on_map(
    polygons: Union[np.ndarray, list], values: np.ndarray, opacity: float, map_color_scale: str, **kwargs: Any
) -> go.Choroplethmap:
    """Plot a set of polygons on a map.

    Args:
        polygons (Union[np.ndarray, list]): Numpy array or list containing the polygons to plot.
        values (np.ndarray): Numpy array consistent with polygons containing the value which is
                             used in coloring the polygons on the map.
        opacity (float): Float between 0 and 1 specifying the opacity of the polygon fill color.
        map_color_scale (str): The string which defines which plotly color scale.
        **kwargs (Any): Additional key word arguments which can be passed on the go.Choroplethmap object
            (will override the default values as specified in this function)

    Returns:
        trace: go.Choroplethmap trace with the colored polygons which can be added to a go.Figure object.

    """
    polygon_id = list(range(values.shape[0]))
    feature_collection = FeatureCollection([Feature(geometry=polygons[idx], id_value=idx) for idx in polygon_id])
    text_box = [
        f"<b>Polygon ID</b>: {counter:d}<br><b>Center (lon, lat)</b>: "
        f"({polygons[counter].centroid.coords[0][0]:.4f}, {polygons[counter].centroid.coords[0][1]:.4f})<br>"
        f"<b>Value</b>: {values[counter]:f}<br>"
        for counter in polygon_id
    ]

    trace_options = {
        "geojson": feature_collection,
        "featureidkey": "id_value",
        "locations": polygon_id,
        "z": values,
        "marker": {"line": {"width": 0}, "opacity": opacity},
        "hoverinfo": "text",
        "text": text_box,
        "name": "Values",
        "colorscale": map_color_scale,
        "colorbar": {"title": "Values"},
        "showlegend": True,
    }

    for key, value in kwargs.items():
        trace_options[key] = value

    trace = go.Choroplethmap(**trace_options)

    return trace


def plot_regular_grid(
    coordinates: LLA,
    values: np.ndarray,
    opacity: float,
    map_color_scale: str,
    tolerance: float = 1e-7,
    unit: str = "kg/hr",
    name="Values",
) -> go.Choroplethmap:
    """Plots a regular grid of LLA data onto a map.

    So long as the input array is regularly spaced, the value of the spacing is found. A set of rectangles are defined
    where the centre of the rectangle is the LLA coordinate.

    Args:
        coordinates (LLA object): A LLA coordinate object containing a set of locations.
        values (np.array): A set of values that correspond to locations specified in the coordinates.
        opacity (float): The opacity of the grid cells when they are plotted.
        map_color_scale (str): The string which defines which plotly color scale should be used when plotting
            the values.
        tolerance (float, optional): Absolute value above which the difference between values is considered significant.
                                     Used to calculate the regular grid of coordinate values. Defaults to 1e-7.
        unit (str, optional): The unit to be added to the colorscale. Defaults to kg/hr.
        name (str, optional): Name for the trace to be used in the color bar as well

    Returns:
        trace (go.Choroplethmap): Trace with the colored polygons which can be added to a go.Figure object.

    """
    _, gridsize_lat = is_regularly_spaced(coordinates.latitude, tolerance=tolerance)
    _, gridsize_lon = is_regularly_spaced(coordinates.longitude, tolerance=tolerance)

    polygons = [
        geometry.box(
            coordinates.longitude[idx] - gridsize_lon / 2,
            coordinates.latitude[idx] - gridsize_lat / 2,
            coordinates.longitude[idx] + gridsize_lon / 2,
            coordinates.latitude[idx] + gridsize_lat / 2,
        )
        for idx in range(coordinates.nof_observations)
    ]

    trace = plot_polygons_on_map(
        polygons=polygons,
        values=values,
        opacity=opacity,
        name=name,
        colorbar={"title": name + "<br>" + unit},
        map_color_scale=map_color_scale,
    )

    return trace


def plot_hexagonal_grid(
    coordinates: LLA,
    values: np.ndarray,
    opacity: float,
    map_color_scale: str,
    num_hexagons: Union[int, None],
    show_positions: bool,
    aggregate_function: Callable = np.sum,
):
    """Plots a set of values into hexagonal bins with respect to the location of the values.

    Any data points that fall within the area of a hexagon are used to perform aggregation and bin the data.
    See: https://plotly.com/python-api-reference/generated/plotly.figure_factory.create_hexbin_mapbox.html

    Args:
        coordinates (LLA object): A LLA coordinate object containing a set of locations.
        values (np.array): A set of values that correspond to locations specified in the coordinates.
        opacity (float): The opacity of the hexagons when they are plotted.
        map_color_scale (str): Colour scale for plotting values.
        num_hexagons (Union[int, None]): The number of hexagons which define the *horizontal* axis of the plot.
        show_positions (bool): A flag to determine whether the original data should be shown alongside
            the binning hexagons.
        aggregate_function (Callable, optional): Function which to apply on the data in each hexagonal bin to aggregate
            the data and visualise the result.

    Returns:
        (go.Figure): A plotly go figure representing the data which was submitted to this function.

    """
    if num_hexagons is None:
        num_hexagons = max(1, np.ceil((np.max(coordinates.longitude) - np.min(coordinates.longitude)) / 0.25))

    coordinates = coordinates.to_lla()

    hex_plot = ff.create_hexbin_mapbox(
        lat=coordinates.latitude,
        lon=coordinates.longitude,
        color=values,
        nx_hexagon=num_hexagons,
        opacity=opacity,
        agg_func=aggregate_function,
        color_continuous_scale=map_color_scale,
        show_original_data=show_positions,
        original_data_marker={"color": "black"},
    )

    return hex_plot


@dataclass
class Plot:
    """Defines the plot class.

    Can be used to generate various figures from model components while storing general settings to get consistent
    figure appearance.

    Attributes:
        figure_dict (dict): Figure dictionary, used as storage using keys to identify the different figures.
        layout (dict, optional): Layout template for plotly figures, used in all figures generated using this class
            instance.

    """

    figure_dict: dict = field(default_factory=dict)
    layout: dict = field(default_factory=dict)

    def __post_init__(self):
        """Using post init to set the default layout, not able to do this in attribute definition/initialization."""
        self.layout = {
            "layout": go.Layout(
                font={"family": "Futura", "size": 20},
                title={"x": 0.5},
                title_font={"size": 30},
                xaxis={"ticks": "outside", "showline": True, "linewidth": 2},
                yaxis={"ticks": "outside", "showline": True, "linewidth": 2},
                legend={
                    "orientation": "v",
                    "yanchor": "middle",
                    "y": 0.5,
                    "xanchor": "right",
                    "x": 1.2,
                    "font": {"size": 14, "color": "black"},
                },
            )
        }

    def show_all(self, renderer="browser"):
        """Show all the figures which are in the figure dictionary.

        Args:
            renderer (str, optional): Default renderer to use when showing the figures.

        """
        for fig in self.figure_dict.values():
            fig.show(renderer=renderer)

    def plot_single_trace(self, object_to_plot: Union[Type[SlabAndSpike], SourceModel, MCMC], **kwargs: Any):
        """Plotting a trace of a single variable.

        Depending on the object to plot it creates a figure which is stored in the figure_dict attribute.
        First it grabs all the specifics needed for the plot and then plots the trace.

        Args:
            object_to_plot (Union[Type[SlabAndSpike], SourceModel, MCMC]): The object from which to plot a variable
            **kwargs (Any): Additional key word arguments, e.g. burn_in, legend_group, show_legend, dict_key, used in
                some specific plots but not applicable to all.

        """
        plot_specifics = create_trace_specifics(object_to_plot=object_to_plot, **kwargs)

        burn_in = kwargs.pop("burn_in", 0)

        fig = go.Figure()
        fig = plot_single_scatter(
            fig=fig,
            x_values=plot_specifics["x_values"],
            y_values=plot_specifics["y_values"],
            color=plot_specifics["color"],
            name=plot_specifics["name"],
            burn_in=burn_in,
        )

        if burn_in > 0:
            fig.add_vline(
                x=burn_in, line_width=3, line_dash="dash", line_color="black", annotation_text=f"\tBurn in: {burn_in}"
            )
        if isinstance(object_to_plot, SlabAndSpike) and isinstance(object_to_plot, SourceModel):
            prior_num_sources_on = round(object_to_plot.emission_rate.shape[0] * object_to_plot.slab_probability, 2)

            fig.add_hline(
                y=prior_num_sources_on,
                line_width=3,
                line_dash="dash",
                line_color="black",
                annotation_text=f"Prior sources 'on': {prior_num_sources_on}",
            )

        if self.layout is not None:
            fig.update_layout(template=self.layout)

        fig.update_layout(title=plot_specifics["title_text"])
        fig.update_xaxes(title_standoff=20, automargin=True, title_text=plot_specifics["x_label"])
        fig.update_yaxes(title_standoff=20, automargin=True, title_text=plot_specifics["y_label"])

        self.figure_dict[plot_specifics["dict_key"]] = fig

    def plot_trace_per_sensor(
        self,
        object_to_plot: Union[ErrorModel, PerSensor, MCMC],
        sensor_object: Union[SensorGroup, Sensor],
        plot_type: str,
        **kwargs: Any,
    ):
        """Plotting a trace of a single variable per sensor.

        Depending on the object to plot it creates a figure which is stored in the figure_dict attribute.
        First it grabs all the specifics needed for the plot and then plots the trace per sensor.

        Args:
            object_to_plot (Union[ErrorModel, PerSensor, MCMC]): The object which to plot a variable from
            sensor_object (Union[SensorGroup, Sensor]): Sensor object associated with the object_to_plot
            plot_type (str): String specifying a line or box plot.
            **kwargs (Any): Additional key word arguments, e.g. burn_in, legend_group, show_legend, dict_key, used in
                some specific plots but not applicable to all.

        """
        if isinstance(sensor_object, Sensor):
            temp = SensorGroup()
            temp.add_sensor(sensor_object)
            sensor_object = deepcopy(temp)
        plot_specifics = create_plot_specifics(
            object_to_plot=object_to_plot, sensor_object=sensor_object, plot_type=plot_type, **kwargs
        )
        burn_in = kwargs.pop("burn_in", 0)

        fig = go.Figure()
        for sensor_idx, sensor_key in enumerate(sensor_object.keys()):
            color_idx = sensor_idx % len(sensor_object.color_map)
            color = sensor_object.color_map[color_idx]

            if plot_specifics["plot_type"] == "line":
                fig = plot_single_scatter(
                    fig=fig,
                    x_values=plot_specifics["x_values"],
                    y_values=plot_specifics["y_values"][sensor_idx, :],
                    color=color,
                    name=sensor_key,
                    burn_in=burn_in,
                )
            elif plot_specifics["plot_type"] == "box":
                fig = plot_single_box(
                    fig=fig,
                    y_values=plot_specifics["y_values"][sensor_idx, burn_in:].flatten(),
                    color=color,
                    name=sensor_key,
                )

        if burn_in > 0 and plot_specifics["plot_type"] == "line":
            fig.add_vline(
                x=burn_in, line_width=3, line_dash="dash", line_color="black", annotation_text=f"\tBurn in: {burn_in}"
            )

        if self.layout is not None:
            fig.update_layout(template=self.layout)

        fig.update_layout(title=plot_specifics["title_text"])
        fig.update_xaxes(title_standoff=20, automargin=True, title_text=plot_specifics["x_label"])
        fig.update_yaxes(title_standoff=20, automargin=True, title_text=plot_specifics["y_label"])

        self.figure_dict[plot_specifics["dict_key"]] = fig

    def plot_fitted_values_per_sensor(
        self,
        mcmc_object: MCMC,
        sensor_object: Union[SensorGroup, Sensor],
        background_model: TemporalBackground = None,
        burn_in: int = 0,
    ):
        """Plot the fitted values from the mcmc object against time, also shows the estimated background when inputted.

        Based on the inputs it plots the results of the mcmc analysis, being the fitted values of the concentration
        measurements together with the 10th and 90th quantile lines to show the goodness of fit of the estimates.

        The created figure is stored in the figure_dict attribute.

        Args:
            mcmc_object (MCMC): MCMC object which contains the fitted values in the store attribute of the object.
            sensor_object (Union[SensorGroup, Sensor]): Sensor object associated with the object_to_plot
            background_model (TemporalBackground, optional): Background model containing the estimated background.
            burn_in (int, optional): Number of burn-in iterations to discard before calculating the quantiles
                and median. Defaults to 0.

        """
        if "y" not in mcmc_object.store:
            raise ValueError("Missing fitted values ('y') in mcmc_store_object")

        if isinstance(sensor_object, Sensor):
            temp = SensorGroup()
            temp.add_sensor(sensor_object)
            sensor_object = deepcopy(temp)

        y_values_overall = mcmc_object.store["y"]
        dict_key = "fitted_values"
        title_text = "Observations and Predicted Model Values Against Time"
        x_label = "Time"
        y_label = "Concentration (ppm)"
        fig = go.Figure()

        for sensor_idx, sensor_key in enumerate(sensor_object.keys()):
            plot_idx = np.array(sensor_object.sensor_index == sensor_idx)

            x_values = sensor_object[sensor_key].time
            y_values = y_values_overall[plot_idx, burn_in:]

            color_idx = sensor_idx % len(sensor_object.color_map)
            color = sensor_object.color_map[color_idx]

            fig = plot_quantiles_from_array(
                fig=fig, x_values=x_values, y_values=y_values, quantiles=[10, 90], color=color, name=sensor_key
            )

        if isinstance(background_model, TemporalBackground):
            fig = plot_quantiles_from_array(
                fig=fig,
                x_values=background_model.time,
                y_values=background_model.bg,
                quantiles=[10, 90],
                color="rgb(186, 186, 186)",
                name="Background",
            )

            fig.for_each_trace(
                lambda trace: (
                    trace.update(showlegend=True, name="Background") if trace.name == "Median for Background" else ()
                ),
            )

        fig = sensor_object.plot_timeseries(fig=fig, color_map=sensor_object.color_map, mode="markers")

        fig.add_annotation(
            x=1,
            y=1.1,
            yref="paper",
            xref="paper",
            xanchor="left",
            yanchor="top",
            font={"size": 12, "color": "#000000"},
            align="left",
            showarrow=False,
            borderwidth=2,
            borderpad=10,
            bgcolor="#ffffff",
            bordercolor="#000000",
            opacity=0.8,
            text=(
                "<b>Point</b>: Real observation<br><b>Line</b>: Predicted Value<br><b>Shading</b>: " + "Quantiles 10-90"
            ),
        )

        if self.layout is not None:
            fig.update_layout(template=self.layout)

        fig.update_layout(title=title_text)
        fig.update_xaxes(title_standoff=20, automargin=True, title_text=x_label)
        fig.update_yaxes(title_standoff=20, automargin=True, title_text=y_label)

        self.figure_dict[dict_key] = fig

    def plot_emission_rate_estimates(self, source_model_object, y_axis_type="linear", **kwargs: Any):
        """Plot the emission rate estimates source model object against MCMC iteration.

        Based on the inputs it plots the results of the mcmc analysis, being the estimated emission rate values for
        each source location together with the total emissions estimate, which is the sum over all source locations.

        The created figure is stored in the figure_dict attribute.

        After the loop over all sources we add an empty trace to have the legend entry and desired legend group
        behaviour.

        Args:
            source_model_object (SourceModel): Source model object which contains the estimated emission rate estimates.
            y_axis_type (str, optional): String to indicate whether the y-axis should be linear of log scale.
            **kwargs (Any): Additional key word arguments, e.g. burn_in, dict_key, used in some specific plots but not
                applicable to all.

        """
        total_emissions = np.nansum(source_model_object.emission_rate, axis=0)
        x_values = np.array(range(total_emissions.size))

        burn_in = kwargs.pop("burn_in", 0)

        dict_key = "estimated_values_plot"
        title_text = "Estimated Values of Sources With Respect to MCMC Iterations"
        x_label = MCMC_ITERATION_NUMBER_LITERAL
        y_label = "Estimated Emission<br>Values (kg/hr)"

        fig = go.Figure()

        fig = plot_single_scatter(
            fig=fig,
            x_values=x_values,
            y_values=total_emissions,
            color="rgb(239, 85, 59)",
            name="Total Site Emissions",
            burn_in=burn_in,
            show_legend=True,
        )

        for source_idx in range(source_model_object.emission_rate.shape[0]):
            y_values = source_model_object.emission_rate[source_idx, :]
            if source_model_object.individual_source_labels[source_idx] is not None:
                source_label = source_model_object.individual_source_labels[source_idx]
            else:
                source_label = f"Source {source_idx}"

            fig = plot_single_scatter(
                fig=fig,
                x_values=x_values,
                y_values=y_values,
                color=RGB_LIGHT_BLUE,
                name=source_label,
                burn_in=burn_in,
                show_legend=False,
                legend_group="Source traces",
            )

        fig = plot_single_scatter(
            fig=fig,
            x_values=np.array([None]),
            y_values=np.array([None]),
            color=RGB_LIGHT_BLUE,
            name="Source traces",
            burn_in=0,
            show_legend=True,
        )

        if burn_in > 0:
            fig.add_vline(
                x=burn_in, line_width=3, line_dash="dash", line_color="black", annotation_text=f"\tBurn in: {burn_in}"
            )

        if self.layout is not None:
            fig.update_layout(template=self.layout)

        fig.add_annotation(
            x=1.05,
            y=1.05,
            yref="paper",
            xref="paper",
            xanchor="left",
            yanchor="top",
            align="left",
            font={"size": 12, "color": "#000000"},
            showarrow=False,
            borderwidth=2,
            borderpad=10,
            bgcolor="#ffffff",
            bordercolor="#000000",
            opacity=0.8,
            text=(
                "<b>Total Site Emissions</b> are<br>the sum of all estimated<br>"
                "emission rates at a given<br>iteration number."
            ),
        )

        fig.update_layout(title=title_text)
        fig.update_xaxes(title_standoff=20, automargin=True, title_text=x_label)
        fig.update_yaxes(title_standoff=20, automargin=True, title_text=y_label)
        if y_axis_type == "log":
            fig.update_yaxes(type="log")
            dict_key = "log_estimated_values_plot"
        elif y_axis_type != "linear":
            raise ValueError(f"Only linear or log y axis type is allowed, {y_axis_type} was currently specified.")

        self.figure_dict[dict_key] = fig

    def create_empty_map_figure(self, dict_key: str = "map_plot") -> None:
        """Creating an empty map figure to use when you want to add additional traces on a map.

        Args:
            dict_key (str, optional): String key for figure dictionary

        """
        self.figure_dict[dict_key] = go.Figure(
            data=go.Scattermap(),
            layout={
                "map_style": "carto-positron",
                "map_center_lat": 0,
                "map_center_lon": 0,
                "map_zoom": 0,
            },
        )

    def plot_values_on_map(
        self, dict_key: str, coordinates: LLA, values: np.ndarray, aggregate_function: Callable = np.sum, **kwargs: Any
    ):
        """Plot values on a map based on coordinates.

        Args:
            dict_key (str): Sting key to use in the figure dictionary
            coordinates (LLA): LLA coordinates to use in plotting the values on the map
            values (np.ndarray): Numpy array of values consistent with coordinates to plot on the map
            aggregate_function (Callable, optional): Function which to apply on the data in each hexagonal bin to
                aggregate the data and visualise the result.
            **kwargs (Any): Additional keyword arguments for plotting behaviour (opacity, map_color_scale, num_hexagons,
                show_positions)

        """
        map_color_scale = kwargs.pop("map_color_scale", "YlOrRd")
        num_hexagons = kwargs.pop("num_hexagons", None)
        opacity = kwargs.pop("opacity", 0.8)
        show_positions = kwargs.pop("show_positions", False)

        latitude_check, _ = is_regularly_spaced(coordinates.latitude)
        longitude_check, _ = is_regularly_spaced(coordinates.longitude)
        if latitude_check and longitude_check:
            self.create_empty_map_figure(dict_key=dict_key)
            trace = plot_regular_grid(
                coordinates=coordinates,
                values=values,
                opacity=opacity,
                map_color_scale=map_color_scale,
                tolerance=1e-7,
                unit="",
            )
            self.figure_dict[dict_key].add_trace(trace)
        else:
            fig = plot_hexagonal_grid(
                coordinates=coordinates,
                values=values,
                opacity=opacity,
                map_color_scale=map_color_scale,
                num_hexagons=num_hexagons,
                show_positions=show_positions,
                aggregate_function=aggregate_function,
            )
            fig.update_layout(map_style="carto-positron")
            self.figure_dict[dict_key] = fig

        center_longitude = np.mean(coordinates.longitude)
        center_latitude = np.mean(coordinates.latitude)
        self.figure_dict[dict_key].update_layout(
            map={"zoom": 10, "center": {"lon": center_longitude, "lat": center_latitude}}
        )

        if self.layout is not None:
            self.figure_dict[dict_key].update_layout(template=self.layout)

    def plot_quantification_results_on_map(
        self,
        model_object: "ELQModel",
        source_model_to_plot_key: str = None,
        bin_size_x: float = 1,
        bin_size_y: float = 1,
        normalized_count_limit: float = 0.005,
        burn_in: int = 0,
        show_summary_results: bool = True,
        show_fixed_source_locations: bool = True,
    ):
        """Function to create a map with the quantification results of the model object.

        This function takes the "SourceModel" object and calculates the statistics for the quantification results.
        It then populates the figure dictionary with three different maps showing the normalized count,
        median emission rate and the inter-quartile range of the emission rate estimates.

        Args:
            model_object (ELQModel): ELQModel object containing the quantification results
            source_model_to_plot_key (str, optional): Key to use in the model_object.components dictionary to access
              the SourceModel object. If None, defaults to "sources_combined".
            bin_size_x (float, optional): Size of the bins in the x-direction. Defaults to 1.
            bin_size_y (float, optional): Size of the bins in the y-direction. Defaults to 1.
            normalized_count_limit (float, optional): Limit for the normalized count to show on the map.
                Defaults to 0.005.
            burn_in (int, optional): Number of burn-in iterations to discard before calculating the statistics.
                Defaults to 0.
            show_summary_results (bool, optional): Flag to show the summary results on the map. Defaults to True.
            show_fixed_source_locations (bool, optional): Flag to show the fixed sources location when present in one
                of the sourcemaps. Defaults to True.

        """
        if source_model_to_plot_key is None:
            source_model_to_plot_key = "sources_combined"

        source_model = model_object.components[source_model_to_plot_key]
        sensor_object = model_object.sensor_object

        source_locations = source_model.all_source_locations
        emission_rates = source_model.emission_rate

        ref_latitude = source_locations.ref_latitude
        ref_longitude = source_locations.ref_longitude
        ref_altitude = source_locations.ref_altitude

        datetime_min_string = sensor_object.time.min().strftime("%d-%b-%Y, %H:%M:%S")
        datetime_max_string = sensor_object.time.max().strftime("%d-%b-%Y, %H:%M:%S")

        result_weighted, _, normalized_count, count_boolean, enu_points, summary_result = (
            calculate_rectangular_statistics(
                emission_rates=emission_rates,
                source_locations=source_locations,
                bin_size_x=bin_size_x,
                bin_size_y=bin_size_y,
                burn_in=burn_in,
                normalized_count_limit=normalized_count_limit,
            )
        )

        polygons = create_lla_polygons_from_xy_points(
            points_array=enu_points,
            ref_latitude=ref_latitude,
            ref_longitude=ref_longitude,
            ref_altitude=ref_altitude,
            boolean_mask=count_boolean,
        )

        if show_summary_results:
            summary_trace = self.create_summary_trace(summary_result=summary_result)

        self.create_empty_map_figure(dict_key="count_map")
        trace = plot_polygons_on_map(
            polygons=polygons,
            values=normalized_count[count_boolean].flatten(),
            opacity=0.8,
            name="normalized_count",
            colorbar={"title": "Normalized Count", "orientation": "h"},
            map_color_scale="Bluered",
        )
        self.figure_dict["count_map"].add_trace(trace)
        self.figure_dict["count_map"].update_layout(
            map_style="carto-positron",
            map={"zoom": 15, "center": {"lon": ref_longitude, "lat": ref_latitude}},
            title=f"Source location probability "
            f"(>={normalized_count_limit}) for "
            f"{datetime_min_string} to {datetime_max_string}",
            font_family="Futura",
            font_size=15,
        )
        sensor_object.plot_sensor_location(self.figure_dict["count_map"])
        self.figure_dict["count_map"].update_traces(showlegend=False)

        adjusted_result_weights = result_weighted.copy()
        adjusted_result_weights[adjusted_result_weights == 0] = np.nan

        median_of_all_emissions = np.nanmedian(adjusted_result_weights, axis=2)

        self.create_empty_map_figure(dict_key="median_map")

        trace = plot_polygons_on_map(
            polygons=polygons,
            values=median_of_all_emissions[count_boolean].flatten(),
            opacity=0.8,
            name="median_emission",
            colorbar={"title": "Median Emission", "orientation": "h"},
            map_color_scale="Bluered",
        )
        self.figure_dict["median_map"].add_trace(trace)
        self.figure_dict["median_map"].update_layout(
            map_style="carto-positron",
            map={"zoom": 15, "center": {"lon": ref_longitude, "lat": ref_latitude}},
            title=f"Median emission rate estimate for {datetime_min_string} to {datetime_max_string}",
            font_family="Futura",
            font_size=15,
        )
        sensor_object.plot_sensor_location(self.figure_dict["median_map"])
        self.figure_dict["median_map"].update_traces(showlegend=False)

        iqr_of_all_emissions = np.nanquantile(a=adjusted_result_weights, q=0.75, axis=2) - np.nanquantile(
            a=adjusted_result_weights, q=0.25, axis=2
        )
        self.create_empty_map_figure(dict_key="iqr_map")

        trace = plot_polygons_on_map(
            polygons=polygons,
            values=iqr_of_all_emissions[count_boolean].flatten(),
            opacity=0.8,
            name="iqr_emission",
            colorbar={"title": "IQR", "orientation": "h"},
            map_color_scale="Bluered",
        )
        self.figure_dict["iqr_map"].add_trace(trace)
        self.figure_dict["iqr_map"].update_layout(
            map_style="carto-positron",
            map={"zoom": 15, "center": {"lon": ref_longitude, "lat": ref_latitude}},
            title=f"Inter Quartile range (25%-75%) of emission rate "
            f"estimate for {datetime_min_string} to {datetime_max_string}",
            font_family="Futura",
            font_size=15,
        )
        sensor_object.plot_sensor_location(self.figure_dict["iqr_map"])
        self.figure_dict["iqr_map"].update_traces(showlegend=False)

        if show_fixed_source_locations:
            for key, _ in model_object.components.items():
                if bool(re.search("fixed", key)):
                    source_model_fixed = model_object.components[key]
                    source_locations_fixed = source_model_fixed.all_source_locations
                    source_location_fixed_lla = source_locations_fixed.to_lla()
                    sources_lat = source_location_fixed_lla.latitude[:, 0]
                    sources_lon = source_location_fixed_lla.longitude[:, 0]
                    fixed_source_location_trace = go.Scattermap(
                        mode="markers",
                        lon=sources_lon,
                        lat=sources_lat,
                        name=f"Fixed source locations, {key}",
                        marker={"size": 10, "opacity": 0.8},
                    )
                    self.figure_dict["count_map"].add_trace(fixed_source_location_trace)
                    self.figure_dict["median_map"].add_trace(fixed_source_location_trace)
                    self.figure_dict["iqr_map"].add_trace(fixed_source_location_trace)

        if show_summary_results:
            self.figure_dict["count_map"].add_trace(summary_trace)
            self.figure_dict["count_map"].update_traces(showlegend=True)
            self.figure_dict["median_map"].add_trace(summary_trace)
            self.figure_dict["median_map"].update_traces(showlegend=True)
            self.figure_dict["iqr_map"].add_trace(summary_trace)
            self.figure_dict["iqr_map"].update_traces(showlegend=True)

    def plot_coverage(
        self,
        coordinates: LLA,
        couplings: np.ndarray,
        threshold_function: Callable = np.max,
        coverage_threshold: float = 6,
        opacity: float = 0.8,
        map_color_scale="jet",
    ):
        """Creates a coverage plot using the coverage function from Gaussian Plume.

        Args:
            coordinates (LLA object): A LLA coordinate object containing a set of locations.
            couplings (np.array): The calculated values of coupling (The 'A matrix') for a set of wind data.
            threshold_function (Callable, optional): Callable function which returns some single value that defines the
                                         maximum or 'threshold' coupling. Examples: np.quantile(q=0.9),
                                         np.max, np.mean. Defaults to np.max.
            coverage_threshold (float, optional): The threshold value of the estimated emission rate which is
                                                  considered to be within the coverage. Defaults to 6 kg/hr.
            opacity (float): The opacity of the grid cells when they are plotted.
            map_color_scale (str): The string which defines which plotly colour scale should be used when plotting
                                   the values.

        """
        coverage_values = GaussianPlume(source_map=None).compute_coverage(
            couplings=couplings, threshold_function=threshold_function, coverage_threshold=coverage_threshold
        )
        self.plot_values_on_map(
            dict_key="coverage_map",
            coordinates=coordinates,
            values=coverage_values,
            aggregate_function=np.max,
            opacity=opacity,
            map_color_scale=map_color_scale,
        )

    @staticmethod
    def create_summary_trace(
        summary_result: pd.DataFrame,
    ) -> go.Scattermap:
        """Helper function to create the summary information to plot on top of map type plots.

        We use the summary result calculated through the support functions module to create a trace which contains
        the summary information for each source location.

        Args:
            summary_result (pd.DataFrame): DataFrame containing the summary information for each source location.

        Returns:
            summary_trace (go.Scattermap): Trace with summary information to plot on top of map type plots.

        """
        summary_text_values = [
            f"<b>Source ID</b>: {value}<br>"
            f"<b>(Lon, Lat, Alt)</b> ([deg], [deg], [m]):<br>"
            f"({summary_result.longitude[value]:.7f}, "
            f"{summary_result.latitude[value]:.7f}, {summary_result.altitude[value]:.3f})<br>"
            f"<b>Height</b>: {summary_result.height[value]:.3f} [m]<br>"
            f"<b>Median emission rate</b>: {summary_result.median_estimate[value]:.4f} [kg/hr]<br>"
            f"<b>2.5% quantile</b>: {summary_result.quantile_025[value]:.3f} [kg/hr]<br>"
            f"<b>97.5% quantile</b>: {summary_result.quantile_975[value]:.3f} [kg/hr]<br>"
            f"<b>IQR</b>: {summary_result.iqr_estimate[value]:.4f} [kg/hr]<br>"
            f"<b>Blob present during</b>: "
            f"{summary_result.absolute_count_iterations[value]:.0f} iterations<br>"
            f"<b>Blob likelihood</b>: {summary_result.blob_likelihood[value]:.5f}<br>"
            for value in summary_result.index
        ]

        summary_trace = go.Scattermap(
            lat=summary_result.latitude,
            lon=summary_result.longitude,
            mode="markers",
            marker=go.scattermap.Marker(size=14, color="black"),
            text=summary_text_values,
            name="Summary",
            hoverinfo="text",
        )

        return summary_trace



# SPDX-FileCopyrightText: 2024 Shell Global Solutions International B.V. All Rights Reserved.
#
# SPDX-License-Identifier: Apache-2.0
"""Support Functions Module."""
__all__ = ["post_processing", "spatio_temporal_interpolation"]



# SPDX-FileCopyrightText: 2024 Shell Global Solutions International B.V. All Rights Reserved.
#
# SPDX-License-Identifier: Apache-2.0

# -*- coding: utf-8 -*-
"""Error model module."""
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Union

import numpy as np
from openmcmc import parameter
from openmcmc.distribution.distribution import Gamma
from openmcmc.model import Model
from openmcmc.sampler.sampler import NormalGamma

from pyelq.component.component import Component
from pyelq.gas_species import GasSpecies
from pyelq.meteorology import MeteorologyGroup
from pyelq.sensor.sensor import Sensor, SensorGroup

if TYPE_CHECKING:
    from pyelq.plotting.plot import Plot


@dataclass
class ErrorModel(Component):
    """Measurement precision model component for the model.

    Attributes:
        n_sensor (int): number of sensors in the sensor object used for analysis.
        precision_index (np.ndarray): index mapping precision parameters onto observations. Will be set up differently
            for different model types.
        precision_parameter (parameter.Parameter): parameter object which constructs the full measurement error
            precision matrix from the components stored in state. Will be passed to the distribution for the observed
            when the full model is constructed.
        prior_precision_shape (Union[np.ndarray, float]): prior shape parameters for the precision model. Set up
            differently per model type.
        prior_precision_rate (Union[np.ndarray, float]): prior rate parameters for the precision model. Set up
            differently per model type.
        initial_precision (Union[np.ndarray, float]): initial value for the precision to be passed to the analysis
            routine. Set up differently per model type.
        precision (np.ndarray): array of sampled measurement error precision values, populated in self.from_mcmc() after
            the MCMC run is completed.

    """

    n_sensor: int = field(init=False)
    precision_index: np.ndarray = field(init=False)
    precision_parameter: parameter.Parameter = field(init=False)
    prior_precision_shape: Union[np.ndarray, float] = field(init=False)
    prior_precision_rate: Union[np.ndarray, float] = field(init=False)
    initial_precision: Union[np.ndarray, float] = field(init=False)
    precision: np.ndarray = field(init=False)

    def initialise(
        self, sensor_object: SensorGroup, meteorology: MeteorologyGroup = None, gas_species: GasSpecies = None
    ):
        """Take data inputs and extract relevant properties.

        Args:
            sensor_object (SensorGroup): sensor data.
            meteorology (MeteorologyGroup): meteorology data. Defaults to None.
            gas_species (GasSpecies): gas species information. Defaults to None.

        """
        self.n_sensor = sensor_object.nof_sensors

    def make_model(self, model: list = None) -> list:
        """Take model list and append new elements from current model component.

        Args:
            model (list, optional): Current list of model elements. Defaults to None.

        Returns:
            list: model output list.

        """
        if model is None:
            model = []
        model.append(Gamma("tau", shape="a_tau", rate="b_tau"))
        return model

    def make_sampler(self, model: Model, sampler_list: list = None) -> list:
        """Take sampler list and append new elements from current model component.

        Args:
            model (Model): Full model list of distributions.
            sampler_list (list, optional): Current list of samplers. Defaults to None.

        Returns:
            list: sampler output list.

        """
        if sampler_list is None:
            sampler_list = []
        sampler_list.append(NormalGamma("tau", model))
        return sampler_list

    def make_state(self, state: dict = None) -> dict:
        """Take state dictionary and append initial values from model component.

        Args:
            state (dict, optional): current state vector. Defaults to None.

        Returns:
            dict: current state vector with components added.

        """
        if state is None:
            state = {}
        state["a_tau"] = self.prior_precision_shape.flatten()
        state["b_tau"] = self.prior_precision_rate.flatten()
        state["precision_index"] = self.precision_index
        state["tau"] = self.initial_precision.flatten()
        return state

    def from_mcmc(self, store: dict):
        """Extract results of mcmc from mcmc.store and attach to components.

        Args:
            store (dict): mcmc result dictionary.

        """
        self.precision = store["tau"]


@dataclass
class BySensor(ErrorModel):
    """Version of measurement precision where each sensor object has a different precision.

    Attributes:
        prior_precision_shape (Union[np.ndarray, float]): prior shape parameters for the precision model, can be
            specified either as a float or as a (nof_sensors, ) np.ndarray: a float specification will result in
            the same parameter value for each sensor. Defaults to 1e-3.
        prior_precision_rate (Union[np.ndarray, float]): prior rate parameters for the precision model, can be
            specified either as a float or as a (nof_sensors, ) np.ndarray: a float specification will result in
            the same parameter value for each sensor. Defaults to 1e-3.
        initial_precision (Union[np.ndarray, float]): initial value for the precision parameters, can be specified
            either as a float or as a (nof_sensors, ) np.ndarray: a float specification will result in the same
            parameter value for each sensor. Defaults to 1.
        precision_index (np.ndarray): index mapping precision parameters onto observations. Parameters 1:n_sensor are
            mapped as the measurement error precisions of the corresponding sensors.
        precision_parameter (Parameter.MixtureParameterMatrix): parameter specification for this model, maps the
            current value of the parameter in the state dict onto the concentration data precisions.

    """

    prior_precision_shape: Union[np.ndarray, float] = 1e-3
    prior_precision_rate: Union[np.ndarray, float] = 1e-3
    initial_precision: Union[np.ndarray, float] = 1.0

    def initialise(
        self, sensor_object: SensorGroup, meteorology: MeteorologyGroup = None, gas_species: GasSpecies = None
    ):
        """Set up the error model using sensor properties.

        Args:
            sensor_object (SensorGroup): sensor data.
            meteorology (MeteorologyGroup): meteorology data. Defaults to None.
            gas_species (GasSpecies): gas species information. Defaults to None.

        """
        super().initialise(sensor_object=sensor_object, meteorology=meteorology, gas_species=gas_species)
        self.prior_precision_shape = self.prior_precision_shape * np.ones((self.n_sensor,))
        self.prior_precision_rate = self.prior_precision_rate * np.ones((self.n_sensor,))
        self.initial_precision = self.initial_precision * np.ones((self.n_sensor,))
        self.precision_index = sensor_object.sensor_index
        self.precision_parameter = parameter.MixtureParameterMatrix(param="tau", allocation="precision_index")

    def plot_iterations(self, plot: "Plot", sensor_object: Union[SensorGroup, Sensor], burn_in_value: int) -> "Plot":
        """Plots the error model values for every sensor with respect to the MCMC iterations.

        Args:
            sensor_object (Union[SensorGroup, Sensor]): Sensor object associated with the error_model
            burn_in_value (int): Burn in value to show in plot.
            plot (Plot): Plot object to which this figure will be added in the figure dictionary

        Returns:
            plot (Plot): Plot object to which this figure is added in the figure dictionary with
                key 'error_model_iterations'

        """
        plot.plot_trace_per_sensor(
            object_to_plot=self, sensor_object=sensor_object, plot_type="line", burn_in=burn_in_value
        )

        return plot

    def plot_distributions(self, plot: "Plot", sensor_object: Union[SensorGroup, Sensor], burn_in_value: int) -> "Plot":
        """Plots the distribution of the error model values after the burn in for every sensor.

        Args:
            sensor_object (Union[SensorGroup, Sensor]): Sensor object associated with the error_model
            burn_in_value (int): Burn in value to show in plot.
            plot (Plot): Plot object to which this figure will be added in the figure dictionary

        Returns:
            plot (Plot): Plot object to which this figure is added in the figure dictionary with
                key 'error_model_distributions'

        """
        plot.plot_trace_per_sensor(
            object_to_plot=self, sensor_object=sensor_object, plot_type="box", burn_in=burn_in_value
        )

        return plot


@dataclass
class ByRelease(ErrorModel):
    """ByRelease error model, special case of the measurement precision model.

    Version of the measurement precision model where each sensor object has a different precision, and there are
    different precisions for periods inside and outside controlled release periods. For all parameters: the first
    element corresponds to the case where the sources are OFF; the second element corresponds to the case where the
    sources are ON.

    Attributes:
        prior_precision_shape (np.ndarray): prior shape parameters for the precision model, can be
            specified either as a (2, 1) np.ndarray or as a (2, nof_sensors) np.ndarray: the former specification
            will result in the same prior specification for the off/on precisions for each sensor. Defaults to
            np.array([1e-3, 1e-3]).
        prior_precision_rate (np.ndarray): prior rate parameters for the precision model, can be
            specified either as a (2, 1) np.ndarray or as a (2, nof_sensors) np.ndarray: the former specification
            will result in the same prior specification for the off/on precisions for each sensor. Defaults to
            np.array([1e-3, 1e-3]).
        initial_precision (np.ndarray): initial value for the precision parameters, can be
            specified either as a (2, 1) np.ndarray or as a (2, nof_sensors) np.ndarray: the former specification
            will result in the same prior specification for the off/on precisions for each sensor. Defaults to
            np.array([1.0, 1.0]).
        precision_index (np.ndarray): index mapping precision parameters onto observations. Parameters 1:n_sensor are
            mapped onto each sensor for the periods where the sources are OFF; parameters (n_sensor + 1):(2 * n_sensor)
            are mapped onto each sensor for the periods where the sources are ON.
        precision_parameter (Parameter.MixtureParameterMatrix): parameter specification for this model, maps the
            current value of the parameter in the state dict onto the concentration data precisions.

    """

    prior_precision_shape: np.ndarray = field(default_factory=lambda: np.array([1e-3, 1e-3], ndmin=2).T)
    prior_precision_rate: np.ndarray = field(default_factory=lambda: np.array([1e-3, 1e-3], ndmin=2).T)
    initial_precision: np.ndarray = field(default_factory=lambda: np.array([1.0, 1.0], ndmin=2).T)

    def initialise(
        self, sensor_object: SensorGroup, meteorology: MeteorologyGroup = None, gas_species: GasSpecies = None
    ):
        """Set up the error model using sensor properties.

        Args:
            sensor_object (SensorGroup): sensor data.
            meteorology (MeteorologyGroup): meteorology data. Defaults to None.
            gas_species (GasSpecies): gas species information. Defaults to None.

        """
        super().initialise(sensor_object=sensor_object, meteorology=meteorology, gas_species=gas_species)
        self.prior_precision_shape = self.prior_precision_shape * np.ones((2, self.n_sensor))
        self.prior_precision_rate = self.prior_precision_rate * np.ones((2, self.n_sensor))
        self.initial_precision = self.initial_precision * np.ones((2, self.n_sensor))
        self.precision_index = sensor_object.sensor_index + sensor_object.source_on * self.n_sensor
        self.precision_parameter = parameter.MixtureParameterMatrix(param="tau", allocation="precision_index")

    def plot_iterations(self, plot: "Plot", sensor_object: Union[SensorGroup, Sensor], burn_in_value: int) -> "Plot":
        """Plot the estimated error model parameters against iterations of the MCMC chain.

        Works by simply creating a separate plot for each of the two categories of precision parameter (when the
        sources are on/off). Creates a BySensor() object for each of the off/on precision cases, and then makes a
        call to its plot function.

        Args:
            sensor_object (Union[SensorGroup, Sensor]): Sensor object associated with the error_model
            burn_in_value (int): Burn in value to show in plot.
            plot (Plot): Plot object to which this figure will be added in the figure dictionary

        Returns:
            plot (Plot): Plot object to which this figure is added in the figure dictionary with
                key 'error_model_iterations'

        """
        figure_keys = ["error_model_off_iterations", "error_model_on_iterations"]
        figure_titles = [
            "Estimated error parameter values: sources off",
            "Estimated error parameter values: sources on",
        ]
        precision_arrays = [
            self.precision[: sensor_object.nof_sensors, :],
            self.precision[sensor_object.nof_sensors :, :],
        ]
        for key, title, array in zip(figure_keys, figure_titles, precision_arrays):
            error_model = BySensor()
            error_model.precision = array
            plot = error_model.plot_iterations(plot, sensor_object, burn_in_value)
            plot.figure_dict[key] = plot.figure_dict.pop("error_model_iterations")
            plot.figure_dict[key].update_layout(title=title)
        return plot

    def plot_distributions(self, plot: "Plot", sensor_object: Union[SensorGroup, Sensor], burn_in_value: int) -> "Plot":
        """Plot the estimated distributions of error model parameters.

        Works by simply creating a separate plot for each of the two categories of precision parameter (when the
        sources are off/on). Creates a BySensor() object for each of the off/on precision cases, and then makes a
        call to its plot function.

        Args:
            sensor_object (Union[SensorGroup, Sensor]): Sensor object associated with the error_model
            burn_in_value (int): Burn in value to show in plot.
            plot (Plot): Plot object to which this figure will be added in the figure dictionary

        Returns:
            plot (Plot): Plot object to which this figure is added in the figure dictionary with
                key 'error_model_distributions'

        """
        figure_keys = ["error_model_off_distributions", "error_model_on_distributions"]
        figure_titles = [
            "Estimated error parameter distribution: sources off",
            "Estimated error parameter distribution: sources on",
        ]
        precision_arrays = [
            self.precision[: sensor_object.nof_sensors, :],
            self.precision[sensor_object.nof_sensors :, :],
        ]
        for key, title, array in zip(figure_keys, figure_titles, precision_arrays):
            error_model = BySensor()
            error_model.precision = array
            plot = error_model.plot_distributions(plot, sensor_object, burn_in_value)
            plot.figure_dict[key] = plot.figure_dict.pop("error_model_distributions")
            plot.figure_dict[key].update_layout(title=title)
        return plot



# SPDX-FileCopyrightText: 2024 Shell Global Solutions International B.V. All Rights Reserved.
#
# SPDX-License-Identifier: Apache-2.0

"""Model components for background modelling."""

from abc import abstractmethod
from copy import deepcopy
from dataclasses import dataclass, field
from typing import Union

import numpy as np
import pandas as pd
from openmcmc import gmrf, parameter
from openmcmc.distribution.distribution import Gamma
from openmcmc.distribution.location_scale import Normal
from openmcmc.model import Model
from openmcmc.sampler.sampler import NormalGamma, NormalNormal
from scipy import sparse
from sklearn.neighbors import NearestNeighbors

from pyelq.component.component import Component
from pyelq.coordinate_system import Coordinate
from pyelq.gas_species import GasSpecies
from pyelq.meteorology import MeteorologyGroup
from pyelq.sensor.beam import Beam
from pyelq.sensor.sensor import SensorGroup


@dataclass
class Background(Component):
    """Superclass for background models.

    Attributes:
        n_obs (int): total number of observations in the background model (across all sensors).
        n_parameter (int): number of parameters in the background model
        bg (np.ndarray): array of sampled background values, populated in self.from_mcmc() after the MCMC run is
            completed.
        precision_scalar (np.ndarray): array of sampled background precision values, populated in self.from_mcmc() after
            the MCMC run is completed. Only populated if update_precision is True.
        precision_matrix (Union[np.ndarray, sparse.csr_array]): un-scaled precision matrix for the background parameter
            vector.
        mean_bg (float): global mean background value. Should be populated from the value specified in the GasSpecies
            object.
        update_precision (bool): logical determining whether the background (scalar) precision parameter should be
            updated as part of the MCMC. Defaults to False.
        prior_precision_shape (float): shape parameter for the prior gamma distribution for the scalar precision
            parameter(s).
        prior_precision_rate (float): rate parameter for the prior gamma distribution for the scalar precision
            parameter(s).
        initial_precision (float): initial value for the scalar precision parameter.
        basis_matrix (sparse.csr_array): [n_obs x n_time] matrix mapping the background model parameters on to the
            observations.

    """

    n_obs: int = field(init=False)
    n_parameter: int = field(init=False)
    bg: np.ndarray = field(init=False)
    precision_scalar: np.ndarray = field(init=False)
    precision_matrix: Union[np.ndarray, sparse.csc_matrix] = field(init=False)
    mean_bg: Union[float, None] = None
    update_precision: bool = False
    prior_precision_shape: float = 1e-3
    prior_precision_rate: float = 1e-3
    initial_precision: float = 1.0
    basis_matrix: sparse.csr_array = field(init=False)

    @abstractmethod
    def initialise(self, sensor_object: SensorGroup, meteorology: MeteorologyGroup, gas_species: GasSpecies):
        """Take data inputs and extract relevant properties.

        Args:
            sensor_object (SensorGroup): sensor data
            meteorology (MeteorologyGroup): meteorology data
            gas_species (GasSpecies): gas species information

        """

    def make_model(self, model: list = None) -> list:
        """Take model list and append new elements from current model component.

        Args:
            model (list, optional): Current list of model elements. Defaults to None.

        Returns:
            list: model output list.

        """
        bg_precision_predictor = parameter.ScaledMatrix(matrix="P_bg", scalar="lambda_bg")
        model.append(Normal("bg", mean="mu_bg", precision=bg_precision_predictor))
        if self.update_precision:
            model.append(Gamma("lambda_bg", shape="a_lam_bg", rate="b_lam_bg"))
        return model

    def make_sampler(self, model: Model, sampler_list: list = None) -> list:
        """Take sampler list and append new elements from current model component.

        Args:
            model (Model): Full model list of distributions.
            sampler_list (list, optional): Current list of samplers. Defaults to None.

        Returns:
            list: sampler output list.

        """
        if sampler_list is None:
            sampler_list = []
        sampler_list.append(NormalNormal("bg", model))
        if self.update_precision:
            sampler_list.append(NormalGamma("lambda_bg", model))
        return sampler_list

    def make_state(self, state: dict = None) -> dict:
        """Take state dictionary and append initial values from model component.

        Args:
            state (dict, optional): current state vector. Defaults to None.

        Returns:
            dict: current state vector with components added.

        """
        state["mu_bg"] = np.ones((self.n_parameter, 1)) * self.mean_bg
        state["B_bg"] = self.basis_matrix
        state["bg"] = np.ones((self.n_parameter, 1)) * self.mean_bg
        state["P_bg"] = self.precision_matrix
        state["lambda_bg"] = self.initial_precision
        if self.update_precision:
            state["a_lam_bg"] = self.prior_precision_shape
            state["b_lam_bg"] = self.prior_precision_rate
        return state

    def from_mcmc(self, store: dict):
        """Extract results of mcmc from mcmc.store and attach to components.

        Args:
            store (dict): mcmc result dictionary.

        """
        self.bg = store["bg"]
        if self.update_precision:
            self.precision_scalar = store["lambda_bg"]


@dataclass
class TemporalBackground(Background):
    """Model which imposes only temporal correlation on the background parameters.

    Assumes that the prior mean concentration of the background at every location/time point is the global average
    background concentration as defined in the input GasSpecies object.

    Generates the (un-scaled) prior background precision matrix using the function gmrf.precision_temporal: this
    precision matrix imposes first-oder Markov structure for the temporal dependence.

    By default, the times used for the model definition are the set of unique times in the observation set.

    This background model only requires the initialise function, and does not require the implementation of any further
    methods.

    Attributes:
        time (Union[np.ndarray, pd.arrays.DatetimeArray]): vector of times used in defining the model.

    """

    time: Union[np.ndarray, pd.arrays.DatetimeArray] = field(init=False)

    def initialise(self, sensor_object: SensorGroup, meteorology: MeteorologyGroup, gas_species: GasSpecies):
        """Create temporal background model from sensor, meteorology and gas species inputs.

        Args:
            sensor_object (SensorGroup): sensor data object.
            meteorology (MeteorologyGroup): meteorology data object.
            gas_species (GasSpecies): gas species data object.

        """
        self.n_obs = sensor_object.nof_observations
        self.time, unique_inverse = np.unique(sensor_object.time, return_inverse=True)
        self.time = pd.array(self.time, dtype="datetime64[ns]")
        self.n_parameter = len(self.time)
        self.basis_matrix = sparse.csr_array((np.ones(self.n_obs), (np.array(range(self.n_obs)), unique_inverse)))
        self.precision_matrix = gmrf.precision_temporal(time=self.time)
        if self.mean_bg is None:
            self.mean_bg = gas_species.global_background


@dataclass
class SpatioTemporalBackground(Background):
    """Model which imposes both spatial and temporal correlation on the background parameters.

    Defines a grid in time, and assumes a correlated time-series per sensor using the defined time grid.

    The background parameter is an [n_location * n_time x 1] (if self.spatial_dependence is True) or an [n_time x 1]
    vector (if self.spatial_dependence is False). In the spatio-temporal case, the background vector is assumed to
    unwrap over space and time as follows:
    bg = [b_1(t_1), b_2(t_1),..., b_nlct(t_1),...,b_1(t_k),..., b_nlct(t_k),...].T
    where nlct is the number of sensor locations.
    This unwrapping mechanism is chosen as it greatly speeds up the sparse matrix operations in the solver (vs. the
    alternative).

    self.basis_matrix is set up to map the elements of the full background vector onto the observations, on the basis
    of spatial location and nearest time knot.

    The temporal background correlation is computed using gmrf.precision_temporal, and the spatial correlation is
    computed using a squared exponential correlation function, parametrized by self.spatial_correlation_param (spatial
    correlation, measured in metres). The full precision matrix is simply a Kronecker product between the two
    component precision matrices.

    Attributes:
        n_time (int): number of time knots for which the model is defined. Note that this does not need to be the same
            as the number of concentration observations in the analysis.
        n_location (int): number of spatial knots in the model.
        time (pd.arrays.DatetimeArray): vector of times used in defining the model.
        spatial_dependence (bool): flag indicating whether the background parameters should be spatially correlated. If
            True, the model assumes a separate background time-series per sensor location, and assumes these
            time-series to be spatially correlated. If False (default), the background parameters are assumed to be
            common between sensors (only temporally correlated).
        spatial_correlation_param (float): correlation length parameter, determining the degree of spatial correlation
            imposed on the background time-series. Units are metres. Assumes equal correlation in all spatial
            directions. Defaults to 1.0.
        location (np.ndarray): [n_location x 3] array of sensor locations, used for calculating the spatial correlation
            between the sensor background values. If self.spatial_dependence is False, this attribute is simply set to
            be the location of the first sensor in the sensor object.
        temporal_precision_matrix (Union[np.ndarray, sparse.csc_matrix]): temporal component of the precision matrix.
            The full model precision matrix is the Kronecker product of this matrix with self.spatial_precision_matrix.
        spatial_precision_matrix (np.ndarray): spatial component of the precision matrix. The full model precision
            matrix is the Kronecker product of this matrix with the self.temporal_precision_matrix. Simply set to 1 if
            self.spatial_dependence is False.
        precision_time_0 (float): precision relating to the first time stamp in the model. Defaults to 0.01.

    """

    n_time: Union[int, None] = None
    n_location: int = field(init=False)
    time: pd.arrays.DatetimeArray = field(init=False)
    spatial_dependence: bool = False
    spatial_correlation_param: float = field(init=False, default=1.0)
    location: Coordinate = field(init=False)
    temporal_precision_matrix: Union[np.ndarray, sparse.csc_matrix] = field(init=False)
    spatial_precision_matrix: np.ndarray = field(init=False)
    precision_time_0: float = field(init=False, default=0.01)

    def initialise(self, sensor_object: SensorGroup, meteorology: MeteorologyGroup, gas_species: GasSpecies):
        """Take data inputs and extract relevant properties.

        Args:
            sensor_object (SensorGroup): sensor data
            meteorology (MeteorologyGroup): meteorology data wind data
            gas_species (GasSpecies): gas species information

        """
        self.make_temporal_knots(sensor_object)
        self.make_spatial_knots(sensor_object)
        self.n_parameter = self.n_time * self.n_location
        self.n_obs = sensor_object.nof_observations

        self.make_precision_matrix()
        self.make_parameter_mapping(sensor_object)

        if self.mean_bg is None:
            self.mean_bg = gas_species.global_background

    def make_parameter_mapping(self, sensor_object: SensorGroup):
        """Create the mapping of parameters onto observations, through creation of the associated basis matrix.

        The background vector unwraps first over the spatial (sensor) location dimension, then over the temporal
        dimension. For more detail, see the main class docstring.

        The data vector in the solver state is assumed to consist of the individual sensor data vectors stacked
        consecutively.

        Args:
            sensor_object (SensorGroup): group of sensor objects.

        """
        nn_object = NearestNeighbors(n_neighbors=1, algorithm="kd_tree").fit(self.time.to_numpy().reshape(-1, 1))
        for k, sensor in enumerate(sensor_object.values()):
            _, time_index = nn_object.kneighbors(sensor.time.to_numpy().reshape(-1, 1))
            basis_matrix = sparse.csr_array(
                (np.ones(sensor.nof_observations), (np.array(range(sensor.nof_observations)), time_index.flatten())),
                shape=(sensor.nof_observations, self.n_time),
            )
            if self.spatial_dependence:
                basis_matrix = sparse.kron(basis_matrix, np.eye(N=self.n_location, M=1, k=-k).T)

            if k == 0:
                self.basis_matrix = basis_matrix
            else:
                self.basis_matrix = sparse.vstack([self.basis_matrix, basis_matrix])

    def make_temporal_knots(self, sensor_object: SensorGroup):
        """Create the temporal grid for the model.

        If self.n_time is not specified, then the model will use the unique set of times from the sensor data.

        If self.n_time is specified, then the model will define a time grid with self.n_time elements.

        Args:
            sensor_object (SensorGroup): group of sensor objects.

        """
        if self.n_time is None:
            self.time = pd.array(np.unique(sensor_object.time), dtype="datetime64[ns]")
            self.n_time = len(self.time)
        else:
            self.time = pd.array(
                pd.date_range(start=np.min(sensor_object.time), end=np.max(sensor_object.time), periods=self.n_time),
                dtype="datetime64[ns]",
            )

    def make_spatial_knots(self, sensor_object: SensorGroup):
        """Create the spatial grid for the model.

        If self.spatial_dependence is False, the code assumes that only a single (arbitrary) location is used, thereby
        eliminating any spatial dependence.

        If self.spatial_dependence is True, a separate but correlated time-series of background parameters is assumed
        for each sensor location.

        Args:
            sensor_object (SensorGroup): group of sensor objects.

        """
        if self.spatial_dependence:
            self.n_location = sensor_object.nof_sensors
            self.get_locations_from_sensors(sensor_object)
        else:
            self.n_location = 1
            self.location = sensor_object[list(sensor_object.keys())[0]].location

    def make_precision_matrix(self):
        """Create the full precision matrix for the background parameters.

        Defined as the Kronecker product of the temporal precision matrix and the spatial precision matrix.

        """
        self.temporal_precision_matrix = gmrf.precision_temporal(time=self.time)
        lam = self.temporal_precision_matrix[0, 0]
        self.temporal_precision_matrix[0, 0] = lam * (2.0 - lam / (self.precision_time_0 + lam))

        if self.spatial_dependence:
            self.make_spatial_precision_matrix()
            self.precision_matrix = sparse.kron(self.temporal_precision_matrix, self.spatial_precision_matrix)
        else:
            self.precision_matrix = self.temporal_precision_matrix
        if (self.n_parameter == 1) and sparse.issparse(self.precision_matrix):
            self.precision_matrix = self.precision_matrix.toarray()

    def make_spatial_precision_matrix(self):
        """Create the spatial precision matrix for the model.

        The spatial precision matrix is simply calculated as the inverse of a squared exponential covariance matrix
        calculated using the sensor locations.

        """
        location_array = self.location.to_array()
        spatial_covariance_matrix = np.exp(
            -(1 / (2 * np.power(self.spatial_correlation_param, 2)))
            * (
                np.power(location_array[:, [0]] - location_array[:, [0]].T, 2)
                + np.power(location_array[:, [1]] - location_array[:, [1]].T, 2)
                + np.power(location_array[:, [2]] - location_array[:, [2]].T, 2)
            )
        )
        self.spatial_precision_matrix = np.linalg.inv(
            spatial_covariance_matrix + (1e-6) * np.eye(spatial_covariance_matrix.shape[0])
        )

    def get_locations_from_sensors(self, sensor_object: SensorGroup):
        """Extract the location information from the sensor object.

        Attaches a Coordinate.ENU object as the self.location attribute, with all the sensor locations stored on the
        same object.

        Args:
            sensor_object (SensorGroup): group of sensor objects.

        """
        self.location = deepcopy(sensor_object[list(sensor_object.keys())[0]].location.to_enu())
        self.location.east = np.full(shape=(self.n_location,), fill_value=np.nan)
        self.location.north = np.full(shape=(self.n_location,), fill_value=np.nan)
        self.location.up = np.full(shape=(self.n_location,), fill_value=np.nan)
        for k, sensor in enumerate(sensor_object.values()):
            if isinstance(sensor, Beam):
                self.location.east[k] = np.mean(sensor.location.to_enu().east, axis=0)
                self.location.north[k] = np.mean(sensor.location.to_enu().north, axis=0)
                self.location.up[k] = np.mean(sensor.location.to_enu().up, axis=0)
            else:
                self.location.east[k] = sensor.location.to_enu().east
                self.location.north[k] = sensor.location.to_enu().north
                self.location.up[k] = sensor.location.to_enu().up



# SPDX-FileCopyrightText: 2024 Shell Global Solutions International B.V. All Rights Reserved.
#
# SPDX-License-Identifier: Apache-2.0

"""Components Module."""
__all__ = ["background", "component", "error_model", "offset", "source_model"]



# SPDX-FileCopyrightText: 2024 Shell Global Solutions International B.V. All Rights Reserved.
#
# SPDX-License-Identifier: Apache-2.0

# -*- coding: utf-8 -*-
"""Sensor module.

The superclass for the sensor classes. This module provides the higher level Sensor and SensorGroup classes. The Sensor
class is a single sensor, the SensorGroup is a dictionary of Sensors. The SensorGroup class is created to deal with the
properties over all sensors together.

"""

from copy import deepcopy
from dataclasses import dataclass, field

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from pandas.arrays import DatetimeArray

from pyelq.coordinate_system import ECEF, ENU, LLA, Coordinate


@dataclass
class Sensor:
    """Defines the properties and methods of the sensor class.

    Attributes:
        label (str, optional): String label for sensor
        time (pandas.arrays.DatetimeArray, optional): Array containing time values associated with concentration
            reading
        location (Coordinate, optional): Coordinate object specifying the observation locations
        concentration (np.ndarray, optional): Array containing concentration values associated with time reading
        source_on (np.ndarray, optional): Array of size nof_observations containing boolean values indicating
            whether a source is on or off for each observation, i.e. we are assuming the sensor can/can't see a source

    """

    label: str = field(init=False)
    time: DatetimeArray = field(init=False, default=None)
    location: Coordinate = field(init=False)
    concentration: np.ndarray = field(default_factory=lambda: np.array([]))
    source_on: np.ndarray = field(init=False, default=None)

    @property
    def nof_observations(self) -> int:
        """Int: Number of observations contained in concentration array."""
        return self.concentration.size

    def plot_sensor_location(self, fig: go.Figure, color=None) -> go.Figure:
        """Plotting the sensor location.

        Args:
            fig (go.Figure): Plotly figure object to add the trace to
            color (`optional`): When specified, the color to be used

        Returns:
            fig (go.Figure): Plotly figure object with sensor location trace added to it

        """
        lla_object = self.location.to_lla()

        marker_dict = {"size": 10, "opacity": 0.8}
        if color is not None:
            marker_dict["color"] = color

        fig.add_trace(
            go.Scattermap(
                mode="markers+lines",
                lat=np.array(lla_object.latitude),
                lon=np.array(lla_object.longitude),
                marker=marker_dict,
                line={"width": 3},
                name=self.label,
            )
        )
        return fig

    def plot_timeseries(self, fig: go.Figure, color=None, mode: str = "markers") -> go.Figure:
        """Timeseries plot of the sensor concentration observations.

        Args:
            fig (go.Figure): Plotly figure object to add the trace to
            color (`optional`): When specified, the color to be used
            mode (str, optional): Mode used for plotting, i.e. markers, lines or markers+lines

        Returns:
            fig (go.Figure): Plotly figure object with sensor concentration timeseries trace added to it

        """
        marker_dict = {"size": 5, "opacity": 1}
        if color is not None:
            marker_dict["color"] = color

        fig.add_trace(
            go.Scatter(
                x=self.time,
                y=self.concentration.flatten(),
                mode=mode,
                marker=marker_dict,
                name=self.label,
                legendgroup=self.label,
            )
        )

        return fig


@dataclass
class SensorGroup(dict):
    """A dictionary containing multiple Sensors.

    This class is used when we want to combine a collection of sensors and be able to store/access overall properties.

    Attributes:
        color_map (list, optional): Default colormap to use for plotting

    """

    color_map: list = field(default_factory=list, init=False)

    def __post_init__(self):
        self.color_map = px.colors.qualitative.Pastel

    @property
    def nof_observations(self) -> int:
        """Int: The total number of observations across all the sensors."""
        return int(np.sum([sensor.nof_observations for sensor in self.values()], axis=None))

    @property
    def concentration(self) -> np.ndarray:
        """np.ndarray: Column vector of concentration values across all sensors, unwrapped per sensor."""
        return np.concatenate([sensor.concentration.flatten() for sensor in self.values()], axis=0)

    @property
    def time(self) -> pd.arrays.DatetimeArray:
        """DatetimeArray: Column vector of time values across all sensors."""
        return pd.array(np.concatenate([sensor.time for sensor in self.values()]), dtype="datetime64[ns]")

    @property
    def location(self) -> Coordinate:
        """Coordinate: Coordinate object containing observation locations from all sensors in the group."""
        location_object = deepcopy(list(self.values())[0].location)
        if isinstance(location_object, ENU):
            attr_list = ["east", "north", "up"]
        elif isinstance(location_object, LLA):
            attr_list = ["latitude", "longitude", "altitude"]
        elif isinstance(location_object, ECEF):
            attr_list = ["x", "y", "z"]
        else:
            raise TypeError(
                f"Location object should be either ENU, LLA or ECEF, while currently it is{type(location_object)}"
            )
        for attr in attr_list:
            setattr(
                location_object,
                attr,
                np.concatenate([np.array(getattr(sensor.location, attr), ndmin=1) for sensor in self.values()], axis=0),
            )
        return location_object

    @property
    def sensor_index(self) -> np.ndarray:
        """np.ndarray: Column vector of integer indices linking concentration observation to a particular sensor."""
        return np.concatenate(
            [np.ones(sensor.nof_observations, dtype=int) * i for i, sensor in enumerate(self.values())]
        )

    @property
    def source_on(self) -> np.ndarray:
        """Column vector of booleans indicating whether sources are expected to be on, unwrapped over sensors.

        Assumes source is on when None is specified for a specific sensor.

        Returns:
            np.ndarray: Source on attribute, unwrapped over sensors.

        """
        overall_idx = np.array([])
        for curr_key in list(self.keys()):
            if self[curr_key].source_on is None:
                temp_idx = np.ones(self[curr_key].nof_observations).astype(bool)
            else:
                temp_idx = self[curr_key].source_on

            overall_idx = np.concatenate([overall_idx, temp_idx])
        return overall_idx.astype(bool)

    @property
    def nof_sensors(self) -> int:
        """Int: Number of sensors contained in the SensorGroup."""
        return len(self)

    def add_sensor(self, sensor: Sensor):
        """Add a sensor to the SensorGroup."""
        self[sensor.label] = sensor

    def plot_sensor_location(self, fig: go.Figure, color_map: list = None) -> go.Figure:
        """Plotting of the locations of all sensors in the SensorGroup.

        Args:
            fig (go.Figure): Plotly figure object to add the trace to
            color_map (list, optional): When specified, the colormap to be used, plotting will cycle through
                the colors

        Returns:
            fig (go.Figure): Plotly figure object with sensor location traces added to it

        """
        if color_map is None:
            color_map = self.color_map

        for i, sensor in enumerate(self.values()):
            color_idx = i % len(color_map)
            fig = sensor.plot_sensor_location(fig, color=color_map[color_idx])

        return fig

    def plot_timeseries(self, fig: go.Figure, color_map: list = None, mode: str = "markers") -> go.Figure:
        """Plotting of the concentration timeseries of all sensors in the SensorGroup.

        Args:
            fig (go.Figure): Plotly figure object to add the trace to
            color_map (list, optional): When specified, the colormap to be used, plotting will cycle through
                the colors
            mode (str, optional): Mode used for plotting, i.e. markers, lines or markers+lines

        Returns:
            fig (go.Figure): Plotly figure object with sensor concentration time series traces added to it

        """
        if color_map is None:
            color_map = self.color_map

        for i, sensor in enumerate(self.values()):
            color_idx = i % len(color_map)
            fig = sensor.plot_timeseries(fig, color=color_map[color_idx], mode=mode)

        return fig



# SPDX-FileCopyrightText: 2024 Shell Global Solutions International B.V. All Rights Reserved.
#
# SPDX-License-Identifier: Apache-2.0

# -*- coding: utf-8 -*-
"""Satellite module.

Subclass of Sensor. Mainly used to accommodate satellite sensor TROPOMI. See:
http://www.tropomi.eu/data-products/methane
: http: //www.tropomi.eu/data-products/methane and http://www.tropomi.eu/data-products/nitrogen-dioxide

"""

from dataclasses import dataclass, field

import numpy as np

from pyelq.sensor.sensor import Sensor


@dataclass
class Satellite(Sensor):
    """Defines Satellite sensor class.

    Attributes:
        qa_value (np.ndarray, optional): Array containing quality values associated with the observations.
        precision (np.ndarray, optional): Array containing precision values associated with the observations.
        precision_kernel (np.ndarray, optional): Array containing precision kernel values associated with the
            observations.
        ground_pixel (np.ndarray, optional): Array containing ground pixels values associated with the observations.
            Ground pixels are indicating the dimension perpendicular to the flight direction.
        scanline (np.ndarray, optional): Array containing scanline values associated with the observations.
            Scanlines are indicating the dimension in the direction of flight.
        orbit (np.ndarray, optional): Array containing orbit values associated with the observations.
        pixel_bounds (np.ndarray, optional): Array containing Polygon features which define the pixel bounds.

    """

    qa_value: np.ndarray = field(init=False)
    precision: np.ndarray = field(init=False)
    precision_kernel: np.ndarray = field(init=False)
    ground_pixel: np.ndarray = field(init=False)
    scanline: np.ndarray = field(init=False)
    orbit: np.ndarray = field(init=False, default=None)
    pixel_bounds: np.ndarray = field(init=False)

    def get_orbits(self) -> np.ndarray:
        """Gets the unique orbits which are present in the data.

        Raises:
            ValueError: When orbits attribute is None

        Returns:
            np.ndarray: Unique orbits present in the data.

        """
        if self.orbit is None:
            raise ValueError("Orbits attribute is None")
        return np.unique(self.orbit)



# SPDX-FileCopyrightText: 2024 Shell Global Solutions International B.V. All Rights Reserved.
#
# SPDX-License-Identifier: Apache-2.0
"""Main pyELQ module."""
__all__ = [
    "component",
    "data_access",
    "dispersion_model",
    "plotting",
    "sensor",
    "support_functions",
    "coordinate_system",
    "dlm",
    "gas_species",
    "meteorology",
    "model",
    "preprocessing",
    "source_map",
]



# SPDX-FileCopyrightText: 2024 Shell Global Solutions International B.V. All Rights Reserved.
#
# SPDX-License-Identifier: Apache-2.0

# -*- coding: utf-8 -*-
"""ELQModel module.

This module provides a class definition for the main functionalities of the codebase, providing the interface with the
openMCMC repo and defining some plotting wrappers.

"""
import re
import warnings
from dataclasses import dataclass, field
from typing import Union

import numpy as np
from openmcmc import parameter
from openmcmc.distribution import location_scale
from openmcmc.mcmc import MCMC
from openmcmc.model import Model

from pyelq.component.background import Background, SpatioTemporalBackground
from pyelq.component.error_model import BySensor, ErrorModel
from pyelq.component.offset import PerSensor
from pyelq.component.source_model import Normal, SourceModel
from pyelq.coordinate_system import ENU
from pyelq.gas_species import GasSpecies
from pyelq.meteorology import Meteorology, MeteorologyGroup
from pyelq.plotting.plot import Plot
from pyelq.sensor.sensor import SensorGroup


@dataclass
class ELQModel:
    """Class for setting up, running, and post-processing the full ELQModel analysis.

    Attributes:
        form (dict): dictionary detailing the form of the predictor for the concentration data. For details of the
            required specification, see parameter.LinearCombinationWithTransform() in the openMCMC repo.
        transform (dict): dictionary detailing transformations applied to the model components. For details of the
            required specification, see parameter.LinearCombinationWithTransform() in the openMCMC repo.
        model (Model): full model specification for the analysis, constructed in self.to_mcmc().
        mcmc (MCMC): MCMC object containing model and sampler specification for the problem. Constructed from the
            other components in self.to_mcmc().
        n_iter (int): number of MCMC iterations to be run.
        n_thin (int): number of iterations to thin by.
        fitted_values (np.ndarray): samples of fitted values (i.e. model predictions for the data) generated during the
            MCMC sampler. Attached in self.from_mcmc().

    """

    form: dict = field(init=False)
    transform: dict = field(init=False)
    model: Model = field(init=False)
    mcmc: MCMC = field(init=False)
    n_iter: int = 1000
    n_thin: int = 1
    fitted_values: np.ndarray = field(init=False)

    def __init__(
        self,
        sensor_object: SensorGroup,
        meteorology: Union[Meteorology, MeteorologyGroup],
        gas_species: GasSpecies,
        background: Background = SpatioTemporalBackground(),
        source_model: Union[list, SourceModel] = Normal(),
        error_model: ErrorModel = BySensor(),
        offset_model: PerSensor = None,
    ):
        """Initialise the ELQModel model.

        Model form is as follows:
        y = A*s + b + d + e
        where:
        - y is the vector of observed concentration data (extracted from the sensor object).
        - A*s is the source contribution (from the source model and dispersion model).
        - b is from the background model.
        - d is from the offset model.
        - e is residual error term and var(e) comes from the error precision model.

        Args:
            sensor_object (SensorGroup): sensor data.
            meteorology (Union[Meteorology, MeteorologyGroup]): meteorology data.
            gas_species (GasSpecies): gas species object.
            background (Background): background model specification. Defaults to SpatioTemporalBackground().
            source_model (Union[list, SourceModel]): source model specification. This can be a list of multiple
            SourceModels or a single SourceModel. Defaults to Normal(). If a single SourceModel is used, it will
            be converted to a list.
            error_model (Precision): measurement precision model specification. Defaults to BySensor().
            offset_model (PerSensor): offset model specification. Defaults to None.

        """
        self.sensor_object = sensor_object
        self.meteorology = meteorology
        self.gas_species = gas_species
        self.components = {
            "background": background,
            "error_model": error_model,
            "offset": offset_model,
        }

        if source_model is not None:
            if not isinstance(source_model, list):
                source_model = [source_model]
            for source in source_model:
                if source.label_string is None:
                    self.components["source"] = source
                else:
                    self.components["source_" + source.label_string] = source

        if error_model is None:
            self.components["error_model"] = BySensor()
            warnings.warn("None is not an allowed type for error_model: resetting to default BySensor model.")
        for key in list(self.components.keys()):
            if self.components[key] is None:
                self.components.pop(key)

    def initialise(self):
        """Take data inputs and extract relevant properties."""
        self.form = {}
        self.transform = {}
        for key, component in self.components.items():

            if "background" in key:
                self.form["bg"] = "B_bg"
                self.transform["bg"] = False
            if re.match("source", key):
                source_component_map = component.map
                self.transform[source_component_map["source"]] = False
                self.form[source_component_map["source"]] = source_component_map["coupling_matrix"]
            if "offset" in key:
                self.form["d"] = "B_d"
                self.transform["d"] = False

            self.components[key].initialise(self.sensor_object, self.meteorology, self.gas_species)

    def to_mcmc(self):
        """Convert the ELQModel specification into an MCMC solver object that can be run.

        Executing the following steps:
            - Initialise the model object with the data likelihood (response distribution for y), and add all the
                associated prior distributions, as specified by the model components.
            - Initialise the state dictionary with the observed sensor data, and add parameters associated with all
                the associated prior distributions, as specified by the model components.
            - Initialise the MCMC sampler objects associated with each of the model components.
            - Create the MCMC solver object, using all of the above information.

        """
        response_precision = self.components["error_model"].precision_parameter
        model = [
            location_scale.Normal(
                "y",
                mean=parameter.LinearCombinationWithTransform(self.form, self.transform),
                precision=response_precision,
            )
        ]

        initial_state = {"y": self.sensor_object.concentration}

        for component in self.components.values():
            model = component.make_model(model)
            initial_state = component.make_state(initial_state)

        self.model = Model(model, response={"y": "mean"})

        sampler_list = []
        for component in self.components.values():
            sampler_list = component.make_sampler(self.model, sampler_list)

        self.mcmc = MCMC(initial_state, sampler_list, self.model, n_burn=0, n_iter=self.n_iter, n_thin=self.n_thin)

    def run_mcmc(self):
        """Run the mcmc function."""
        self.mcmc.run_mcmc()

    def from_mcmc(self):
        """Extract information from MCMC solver class once its has run.

        Performs two operations:
            - For each of the components of the model: extracts the related sampled parameter values and attaches these
                to the component class.
            - For all keys in the mcmc.store dictionary: extracts the sampled parameter values from self.mcmc.store and
                puts them into the equivalent fields in the state

        """
        state = self.mcmc.state
        for component in self.components.values():
            component.from_mcmc(self.mcmc.store)
        for key in self.mcmc.store:
            state[key] = self.mcmc.store[key]

        self.make_combined_source_model()

    def make_combined_source_model(self):
        """Aggregate multiple individual source models into a single combined source model.

        This function iterates through the existing source models stored in `self.components` and consolidates them
        into a unified source model named `"sources_combined"`. This is particularly useful when multiple source
        models are involved in an analysis, and a merged representation is required for visualization.

        The combined source model is created as an instance of the `Normal` model, with the label string
        "sources_combined" with the following attributes:
        - emission_rate: concatenated across all source models.
        - all_source_locations: concatenated across all source models.
        - number_on_sources: derived by summing the individual source counts across all source models
        - label_string: concatenated across all source models.
        - individual_source_labels: concatenated across all source models.

        Once combined, the `"sources_combined"` model is stored in the `self.components` dictionary for later use.

        Raises:
            ValueError: If the reference locations of the individual source models are inconsistent.
            This is checked by comparing the reference latitude, longitude, and altitude of each source model.

        """
        combined_model = Normal(label_string="sources_combined")
        emission_rate = np.empty((0, self.mcmc.n_iter))
        all_source_locations_east = np.empty((0, self.mcmc.n_iter))
        all_source_locations_north = np.empty((0, self.mcmc.n_iter))
        all_source_locations_up = np.empty((0, self.mcmc.n_iter))
        number_on_sources = np.empty((0, self.mcmc.n_iter))
        label_string = []
        individual_source_labels = []

        ref_latitude = None
        ref_longitude = None
        ref_altitude = None
        for key, component in self.components.items():
            if re.match("source", key):
                comp_ref_latitude = component.all_source_locations.ref_latitude
                comp_ref_longitude = component.all_source_locations.ref_longitude
                comp_ref_altitude = component.all_source_locations.ref_altitude
                if ref_latitude is None and ref_longitude is None and ref_altitude is None:
                    ref_latitude = comp_ref_latitude
                    ref_longitude = comp_ref_longitude
                    ref_altitude = comp_ref_altitude
                else:
                    if (
                        not np.isclose(ref_latitude, comp_ref_latitude)
                        or not np.isclose(ref_longitude, comp_ref_longitude)
                        or not np.isclose(ref_altitude, comp_ref_altitude)
                    ):
                        raise ValueError(
                            f"Inconsistent reference locations in component '{key}'. "
                            "All source models must share the same reference location."
                        )
                emission_rate = np.concatenate((emission_rate, component.emission_rate))
                number_on_sources = np.concatenate(
                    (
                        number_on_sources.reshape((-1, self.mcmc.n_iter)),
                        component.number_on_sources.reshape(-1, self.mcmc.n_iter),
                    ),
                    axis=0,
                )
                label_string.append(component.label_string)
                individual_source_labels.append(component.individual_source_labels)

                all_source_locations_east = np.concatenate(
                    (
                        all_source_locations_east,
                        component.all_source_locations.east.reshape((-1, self.mcmc.n_iter)),
                    ),
                    axis=0,
                )
                all_source_locations_north = np.concatenate(
                    (
                        all_source_locations_north,
                        component.all_source_locations.north.reshape((-1, self.mcmc.n_iter)),
                    ),
                    axis=0,
                )
                all_source_locations_up = np.concatenate(
                    (
                        all_source_locations_up,
                        component.all_source_locations.up.reshape((-1, self.mcmc.n_iter)),
                    ),
                    axis=0,
                )

        combined_model.all_source_locations = ENU(
            ref_altitude=ref_altitude,
            ref_latitude=ref_latitude,
            ref_longitude=ref_longitude,
            east=all_source_locations_east,
            north=all_source_locations_north,
            up=all_source_locations_up,
        )

        combined_model.emission_rate = emission_rate
        combined_model.label_string = label_string
        combined_model.number_on_sources = np.sum(number_on_sources, axis=0)
        combined_model.individual_source_labels = [item for sublist in individual_source_labels for item in sublist]
        self.components["sources_combined"] = combined_model

    def plot_log_posterior(self, burn_in_value: int, plot: Plot = Plot()) -> Plot():
        """Plots the trace of the log posterior over the iterations of the MCMC.

        Args:
            burn_in_value (int): Burn in value to show in plot.
            plot (Plot, optional): Plot object to which this figure will be added in the figure dictionary

        Returns:
            plot (Plot): Plot object to which this figure is added in the figure dictionary with
                key 'log_posterior_plot'

        """
        plot.plot_single_trace(object_to_plot=self.mcmc, burn_in=burn_in_value)
        return plot

    def plot_fitted_values(self, plot: Plot = Plot()) -> Plot:
        """Plot the fitted values from the mcmc object against time, also shows the estimated background when possible.

        Based on the inputs it plots the results of the mcmc analysis, being the fitted values of the concentration
        measurements together with the 10th and 90th quantile lines to show the goodness of fit of the estimates.

        Args:
            plot (Plot, optional): Plot object to which this figure will be added in the figure dictionary

        Returns:
            plot (Plot): Plot object to which this figure is added in the figure dictionary with key 'fitted_values'

        """
        plot.plot_fitted_values_per_sensor(
            mcmc_object=self.mcmc, sensor_object=self.sensor_object, background_model=self.components["background"]
        )
        return plot
